# Defines external shear profiles
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LensModel.Profiles module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
from herculens.Util import param_util


__all__ = ['Shear', 'ShearGammaPsi']


class Shear(object):
    """External shear in terms of gamma1 and gamma2."""
    param_names = ['gamma1', 'gamma2', 'ra_0', 'dec_0']
    lower_limit_default = {'gamma1': -0.5, 'gamma2': -0.5, 'ra_0': -100, 'dec_0': -100}
    upper_limit_default = {'gamma1': 0.5, 'gamma2': 0.5, 'ra_0': 100, 'dec_0': 100}
    fixed_default = {'gamma1': False, 'gamma2': False, 'ra_0': True, 'dec_0': True}

    def function(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: lensing potential
        """
        x_ = x - ra_0
        y_ = y - dec_0
        f_ = 0.5 * (gamma1 * x_ * x_ + 2 * gamma2 * x_ * y_ - gamma1 * y_ * y_)
        return f_

    def derivatives(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: deflection angles
        """
        x_ = x - ra_0
        y_ = y - dec_0
        f_x = gamma1 * x_ + gamma2 * y_
        f_y = gamma2 * x_ - gamma1 * y_
        return f_x, f_y

    def hessian(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: f_xx, f_yy, f_xy
        """
        gamma1 = gamma1
        gamma2 = gamma2
        kappa = 0
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_yy, f_xy


class ShearGammaPsi(object):
    """External shear in terms of magnitude and position angle."""
    param_names = ['gamma_ext', 'psi_ext', 'ra_0', 'dec_0']
    lower_limit_default = {'gamma_ext': 0, 'psi_ext': -np.pi, 'ra_0': -100, 'dec_0': -100}
    upper_limit_default = {'gamma_ext': 1, 'psi_ext': np.pi, 'ra_0': 100, 'dec_0': 100}
    fixed_default = {'gamma_ext': False, 'psi_ext': False, 'ra_0': True, 'dec_0': True}
    
    def __init__(self):
        self._shear_e1e2 = Shear()
        super(ShearGammaPsi, self).__init__()

    @staticmethod
    def function(x, y, gamma_ext, psi_ext, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma_ext: shear strength
        :param psi_ext: shear angle (radian)
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return:
        """
        # change to polar coordinate
        r, phi = param_util.cart2polar(x-ra_0, y-dec_0)
        f_ = 1. / 2 * gamma_ext * r ** 2 * np.cos(2 * (phi - psi_ext))
        return f_

    def derivatives(self, x, y, gamma_ext, psi_ext, ra_0=0, dec_0=0):
        # rotation angle
        gamma1, gamma2 = param_util.shear_polar2cartesian(psi_ext, gamma_ext)
        return self._shear_e1e2.derivatives(x, y, gamma1, gamma2, ra_0, dec_0)

    def hessian(self, x, y, gamma_ext, psi_ext, ra_0=0, dec_0=0):
        gamma1, gamma2 = param_util.shear_polar2cartesian(psi_ext, gamma_ext)
        return self._shear_e1e2.hessian(x, y, gamma1, gamma2, ra_0, dec_0)



# NOTE: this file is useful for backward-compatibility only, 
# as the preferred way is now to pass the profile class directly 
# to the LightModel() constructor.

from herculens.LightModel.Profiles.sersic import (Sersic, SersicElliptic)
from herculens.LightModel.Profiles.multipole import Multipole
from herculens.LightModel.Profiles.gaussian import (Gaussian, GaussianEllipse)
from herculens.LightModel.Profiles.pixelated import Pixelated
from herculens.LightModel.Profiles.uniform import Uniform
from herculens.LightModel.Profiles.shapelets import Shapelets


# mapping between the string name to the mass profile class.
STRING_MAPPING = {
    'SERSIC': Sersic,
    'SERSIC_ELLIPSE': SersicElliptic,
    'SERSIC_SUPERELLIPSE': SersicElliptic,
    'GAUSSIAN': Gaussian,
    'GAUSSIAN_ELLIPSE': GaussianEllipse,
    'MULTIPOLE': Multipole,
    'PIXELATED': Pixelated,
    'UNIFORM': Uniform,
    'SHAPELETS': Shapelets
}

SUPPORTED_MODELS = list(STRING_MAPPING.keys())



# Handles coordinate grids and convolutions
#
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the ImSim.Numerics module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp
from jax.scipy.ndimage import map_coordinates
from scipy.ndimage import map_coordinates as map_coordinates_orig
from herculens.LensImage.Numerics.grid import RegularGrid
from herculens.LensImage.Numerics.convolution import (PixelKernelConvolution,
                                                      SubgridKernelConvolution,
                                                      GaussianConvolution)
from herculens.Util import kernel_util, util


__all__ = ['Numerics']


class Numerics(object):
    """
    This class manages the numerical options and computations of an image.
    The class has two main functions, re_size_convolve() and coordinates_evaluate()
    """
    def __init__(self, pixel_grid, psf, supersampling_factor=1, convolution_type='jax_scipy_fft',
                 supersampling_convolution=False, iterative_kernel_supersampling=True,
                 supersampling_kernel_size=5, point_source_supersampling_factor=1,
                 convolution_kernel_size=None, truncation=4):
        """

        :param pixel_grid: PixelGrid() class instance
        :param psf: PSF() class instance
        :param supersampling_factor: int, factor of higher resolution sub-pixel sampling of surface brightness
        :param supersampling_convolution: bool, if True, performs (part of) the convolution on the super-sampled
        grid/pixels
        :param point_source_supersampling_factor: super-sampling resolution of the point source placing
        :param convolution_kernel_size: int, odd number, size of convolution kernel. If None, takes size of point_source_kernel
        """
        # if no super sampling, turn the supersampling convolution off
        self._psf_type = psf.psf_type
        if not isinstance(supersampling_factor, int):
            raise TypeError('supersampling_factor needs to be an integer! Current type is %s' % type(supersampling_factor))
        if supersampling_factor == 1:
            supersampling_convolution = False
        self._psf = psf

        self._pixel_width = pixel_grid.pixel_width
        nx, ny = pixel_grid.num_pixel_axes
        transform_pix2angle = pixel_grid.transform_pix2angle
        ra_at_xy_0, dec_at_xy_0 = pixel_grid.radec_at_xy_0
        self._grid = RegularGrid(nx, ny, transform_pix2angle, ra_at_xy_0, dec_at_xy_0, supersampling_factor)
        self._pixel_grid = pixel_grid

        if self._psf_type == 'PIXEL':
            if supersampling_convolution is True:
                kernel_super = psf.kernel_point_source_supersampled(supersampling_factor,
                                                                    iterative_supersampling=iterative_kernel_supersampling)
                if convolution_kernel_size is not None:
                    kernel_super = self._supersampling_cut_kernel(kernel_super, convolution_kernel_size,
                                                                  supersampling_factor)
                self._conv = SubgridKernelConvolution(kernel_super, supersampling_factor,
                                                      supersampling_kernel_size=supersampling_kernel_size)
            else:
                kernel = psf.kernel_point_source
                kernel = self._supersampling_cut_kernel(kernel, convolution_kernel_size,
                                                        supersampling_factor=1)
                self._conv = PixelKernelConvolution(kernel, convolution_type=convolution_type,
                                                    output_shape=(nx, ny))
        elif self._psf_type == 'GAUSSIAN':
            pixel_scale = pixel_grid.pixel_width
            sigma = util.fwhm2sigma(psf.fwhm)
            self._conv = GaussianConvolution(sigma, pixel_scale, supersampling_factor,
                                             supersampling_convolution, truncation=truncation)
        elif self._psf_type == 'NONE':
            self._conv = None
        else:
            raise ValueError('psf_type %s not valid! Chose either NONE, GAUSSIAN or PIXEL.' % self._psf_type)

        if supersampling_convolution is True:
            self._high_res_return = True
        else:
            self._high_res_return = False

        self._point_source_supersampling_factor = point_source_supersampling_factor

    def re_size_convolve(self, flux_array, unconvolved=False, input_as_list=False):
        """
        Resize and convolve the flux array.

        Parameters
        ----------
        flux_array : 1D array
            Flux values corresponding to the coordinates being evaluated.
        unconvolved : bool, optional
            If True, returns the unconvolved image. Default is False.
        input_as_list : bool, optional
            If True, treats the input is a list of flux arrays, and returns
            the resized and convolved flux as a list as well. Default is False.

        Returns
        -------
        image_conv : 2D array
            Convolved image on the regular pixel grid.
        """
        if input_as_list is False:
            return self._re_size_convolve_sgl(flux_array, unconvolved=unconvolved)
        return [
            self._re_size_convolve_sgl(flux_array[i], unconvolved=unconvolved) for i in range(len(flux_array))
        ]
    
    def _re_size_convolve_sgl(self, flux_array, unconvolved=False):
        """

        :param flux_array: 1d array, flux values corresponding to coordinates_evaluate
        :param array_low_res_partial: regular sampled surface brightness, 1d array
        :return: convolved image on regular pixel grid, 2d array
        """
        # add supersampled region to lower resolution on
        image_low_res, image_high_res_partial = self._grid.flux_array2image_low_high(flux_array, high_res_return=self._high_res_return)
        if unconvolved is True or self._conv is None:
            image_conv = image_low_res
        else:
            # convolve low res grid and high res grid
            image_conv = self._conv.re_size_convolve(image_low_res, image_high_res_partial)
        return image_conv * self._pixel_width**2

    def render_point_sources(self, theta_x, theta_y, amplitude):
        """Put the PSF at the locations of multiply imaged point sources.

        Parameters
        ----------
        theta_x : 1D array
            RA of point source positions in the image plane.
        theta_y : 1D array
            Dec of point source positions in the image plane.
        amplitude : 1D array
            Amplitudes of lensed point sources.

        Returns
        -------
        out : 2D array
            Image at the pixel grid resolution where the (interpolated) PSF has
            been placed at the locations of the point sources.

        """
        # TODO Account for supersampling
        result = jnp.zeros(self._pixel_grid.num_pixel_axes)

        # Verify inputs
        theta_x = jnp.atleast_1d(theta_x)
        theta_y = jnp.atleast_1d(theta_y)
        amplitude = jnp.atleast_1d(amplitude)

        # Pixel positions of point sources in the image plane
        x, y = self._pixel_grid.map_coord2pix(theta_x, theta_y)

        # PSF kernel
        if self._psf.kernel_point_source is None:
            err_msg = ("PSF has no kernel_point_source. This can happen, for " +
                "example, if `pixel_size` was not provided for type GAUSSIAN.")
            raise ValueError(err_msg)

        kernel = self._psf.kernel_point_source.T  # taking the transpose for map_coordinates
        nx, ny = self._pixel_grid.num_pixel_axes
        xrange = jnp.arange(nx) + kernel.shape[0] // 2
        yrange = jnp.arange(ny) + kernel.shape[1] // 2

        for x0, y0, amp in zip(x, y, amplitude):
            xy_grid = jnp.meshgrid(xrange - x0, yrange - y0)
            result += amp * map_coordinates(kernel, xy_grid, order=1)

        return result

    @property
    def grid_supersampling_factor(self):
        """

        :return: supersampling factor set for higher resolution sub-pixel sampling of surface brightness
        """
        return self._grid.supersampling_factor

    @property
    def coordinates_evaluate(self):
        """

        :return: 1d array of all coordinates being evaluated to perform the image computation
        """
        return self._grid.coordinates_evaluate

    @staticmethod
    def _supersampling_cut_kernel(kernel_super, convolution_kernel_size, supersampling_factor):
        """

        :param kernel_super: super-sampled kernel
        :param convolution_kernel_size: size of convolution kernel in units of regular pixels (odd)
        :param supersampling_factor: super-sampling factor of convolution kernel
        :return: cut out kernel in super-sampling size
        """
        if convolution_kernel_size is not None:
            size = convolution_kernel_size * supersampling_factor
            if size % 2 == 0:
                size += 1
            kernel_cut = kernel_util.cut_psf(kernel_super, size)
            return kernel_cut
        else:
            return kernel_super

    @property
    def convolution_class(self):
        """

        :return: convolution class (can be SubgridKernelConvolution, PixelKernelConvolution or GaussianConvolution)
        """
        return self._conv

    @property
    def grid_class(self):
        """

        :return: grid class
        """
        return self._grid



# This module defines a generic pixelated model that can be used in conjunction with LightModel or MassModel's pixelated profiles.
# This is a forward model based on Gaussian processes, called a correlated field.
# The implementation is a wrapper around the NIFTy.re (a.k.a JIFTy) correlated field model.

__author__ = 'aymgal'

import numpy as np
import jax
import jax.numpy as jnp
from herculens.Util import jifty_util


__all__ = ['CorrelatedField']


class CorrelatedField(object):
    """Initialize the CorrelatedField class, based on the NIFTy.re correlated field model.
    
    For more details about the `prior_` arguments, please refer to 
    the NIFTy documentation. In particular, this documentation webpage gives
    a visual intuition of the different parameters:
    https://ift.pages.mpcdf.de/nifty/user/old_nifty_getting_started_4_CorrelatedFields.html

    WARNING: the flexibility and asperity parameters have not been thoroughly tested!
    Use these with caution. Moreover, flexibility and asperity are not properly implemented along the spectral direction.

    Parameters
    ----------
    param_prefix : str
        The suffix to be added to the field parameters name.
    mass_or_light_model : object
        The mass or light model instance.
    offset_mean : float, optional
        The global additive offset applied to the field realizations, by default np.log(1e-2).
        NOTE: if `exponentiate` is True, this value should be the log-space of the chosen offset value.
    prior_offset_std : tuple, optional
        The The mean and scatter of the log-normal of the offset, by default (0.5, 1e-6).
    correlation_type : str, optional
        Either the amplitude spectrum ('amplitude') or the power spectrum ('power') of 
        the field are parameterized by the parameters below. By default 'amplitude'.
        The power spectrum is just the amplitude spectrum squared.
    kernel_type : str, optional
        Type of kernel to use for the field, either 'powerlaw' or 'matern', by default 'powerlaw'.
        If 'matern', use the `prior_matern_...` parameters below instead.
    prior_loglogavgslope : tuple, optional
        The mean and scatter of the log-normal distribution for the average slope
         of the power-spectrum in log-log space, by default (-4., 0.5).
    prior_fluctuations : tuple, optional
        The mean and scatter of the log-normal distribution for the fluctuations, by default (1.5, 0.8).
    prior_flexibility : object, optional
        The mean and scatter of the log-normal distribution for the flexibility, by default None.
    prior_asperity : object, optional
        The mean and scatter of the log-normal distribution for the asperity, by default None.
        Can only be set if prior_flexibility is also set.
    prior_matern_scale : tuple, optional
        The mean and scatter of the log-normal distribution for the scale of the Matern kernel, by default (1.0, 0.5).  
    prior_matern_loglogslope : tuple, optional
        The mean and scatter of the log-normal distribution for the log-log slope of the Matern kernel, by default (-4., 0.5).
    prior_matern_cutoff : tuple, optional
        The mean and scatter of the log-normal distribution for the cutoff of the Matern kernel, by default (-10., 1.).
    cropped_border_size : int, optional
        The field can optionally be evaluated on a larger grid size,
        and then cropped to return the model in direct space, by default 0.
        This is the number of pixels added on each size of the pixelated grid.
    num_pix_wl : int, optional
        The number of spectral bands, by default 1.
    correlation_type_wl : str, optional
        Similar to correlation_type. Either the amplitude spectrum ('amplitude') 
        or the power spectrum ('power') of the field are parameterized by the 
        parameters below, or no correlation at all ('none'). By default 'none'.
    prior_loglogavgslope_wl : tuple, optional
        The mean and scatter of the log-normal distribution for the average slope
         of the power-spectrum in log-log space along the spectral dimensions, by default (-2., 0.5).
    prior_fluctuations_wl : tuple, optional
        The mean and scatter of the log-normal distribution for the fluctuations along the spectral dimensions, by default (0.3, 0.8).
    prior_flexibility_wl : object, optional
        The mean and scatter of the log-normal distribution for the flexibility along the spectral dimensions, by default None.
    prior_asperity_wl : object, optional
        The mean and scatter of the log-normal distribution for the asperity along the spectral dimensions, by default None. Can only be set if prior_flexibility_wl is also set.
    num_wl : int, optional
        The number of spectral bands, by default 1.
    cropped_border_size_wl : int, optional
        The field can optionally be evaluated on more than the targeted spectral dimensions,
        and then cropped to return the model in direct space, by default 0.
        This is the number of pixels added on each size of the pixelated grid along the spectral dimensions.
    exponentiate : bool, optional
        Whether to return the exponential of the field, by default True. 
        For example, taking the exponential is useful to ensures non–negative values.

    Raises
    ------
    ValueError
        If the model does not have at least one Pixelated profile.
    ValueError
        If the number of pixels has not been set at creation of the LightModel or MassModel instance.
    """
    def __init__(
            self, 
            param_prefix, 
            mass_or_light_model, 

            # General parameters
            offset_mean=np.log(1e-2),
            prior_offset_std=(0.5, 1e-6),
            correlation_type='amplitude',
            kernel_type='powerlaw',
            cropped_border_size=0,

            # Parameters along the spatial dimensions (`xy_dim`) for kernel_type='powerlaw'
            prior_loglogavgslope=(-4., 0.5), 
            prior_fluctuations=(1.5, 0.8),
            prior_flexibility=None,
            prior_asperity=None,

            # Parameters along the spatial dimensions (`xy_dim`) for kernel_type='matern'
            prior_matern_scale=(0.5, 0.1),
            prior_matern_loglogslope=(-8., 0.5), 
            prior_matern_cutoff=(2., 0.5), 
            matern_renorm_amp=False,

            # Parameters along the spectral, or wavelength dimensions (`wl_dim`)
            num_pix_wl=1,
            correlation_type_wl='none',
            prior_loglogavgslope_wl=(-4., 0.5), 
            prior_fluctuations_wl=(0.3, 0.8),
            prior_flexibility_wl=None,
            prior_asperity_wl=None,
            cropped_border_size_wl=0,

            # Non-linearity
            exponentiate=True,
        ):
        # Sanity checks
        if num_pix_wl < 1:
            raise ValueError("The number of spectral dimensions must be at least 1.")
        # Check that the model is pixelated
        if not mass_or_light_model.has_pixels:
            raise ValueError("The provided LightModel or MassModel must contain "
                             "at least one Pixelated profile for proper use "
                             "with the CorrelatedField model.")
        
        # retrieve the number of pixels from the LightModel instance
        # NOTE: the following is to support both when the CorrelatedField is instantiated
        # before or after the LightModel or a MassModel is passed a LensImage instance.
        # TODO: simplify this when pixelated profiles treatment is improved in LensImage.
        try:
            self._num_pix, self._num_pix_y = mass_or_light_model.pixel_grid.num_pixel_axes
        except AttributeError:
            self._num_pix = mass_or_light_model.pixel_grid_settings.get('num_pixels', None)
        else:
            if self._num_pix != self._num_pix_y:
                raise NotImplementedError("Only square pixel grids are supported for now.")
            
        # Define the 'type' of correlated field
        self._num_pix_wl = num_pix_wl
        if self._num_pix_wl == 1:
            # correlated field along the spatial dimensions only
            self._field_type = '2d'
        elif correlation_type_wl != 'none':
            # correlated field along both spatial and spectral dimensions
            self._field_type = '3d'
        else:
            # stack of 2D correlated fields (i.e. uncorrelated along the spectral dimension)
            self._field_type = '2d_stack'

        # Set the type of correlation kernel
        self._kernel_type = kernel_type
        
        # Pack the prior choices
        if any([p is None for p in [offset_mean, prior_offset_std, prior_loglogavgslope, prior_fluctuations]]):
            raise ValueError("Field parameters and priors `offset_mean`, `prior_offset_std`, `prior_loglogavgslope`, `prior_fluctuations` are mandatory.")
        self._kw_amplitude_offset = {
            'offset_mean': offset_mean,
            'offset_std': prior_offset_std,
        }

        if kernel_type == 'powerlaw':
            self._kw_fluctuations = {
                # Amplitude of field fluctuations
                'fluctuations': prior_fluctuations,

                # Exponent of power law power spectrum component
                'loglogavgslope': prior_loglogavgslope,

                # Extra degrees of freedom
                # NOTE: I did not test much these two in a strong lensing context
                'flexibility': prior_flexibility, 
                'asperity': prior_asperity,
            }
        elif kernel_type == 'matern':
            self._kw_fluctuations = {
                # Matern kernel parameters
                'scale': prior_matern_scale,
                'loglogslope': prior_matern_loglogslope,
                'cutoff': prior_matern_cutoff,
            }

        if self._field_type == '3d':
            self._kw_fluctuations_wl = {
                'fluctuations': prior_fluctuations_wl,
                'loglogavgslope': prior_loglogavgslope_wl,
                'flexibility': prior_flexibility_wl,
                'asperity': prior_asperity_wl,
            }
        else:
            self._kw_fluctuations_wl = None

        # Setup the correlated field model
        self._param_prefix = param_prefix
        self._field_key = 'field'
        self._key_xy = 'xy_dim'
        self._key_wl = 'wl_dim'

        # There are three possibilities:
        # - either it's a 2D correlated field (only along spatial dimensions)
        # - either it's a 3D field, correlated along both spatial and spectral dimensions
        # - either it's a stack of 2D correlated fields (i.e. uncorrelated along the spectral dimension)
        if self._field_type in ('2d', '3d'):
            (
                self._cfm, self._jft_model, 
                self._num_pix_tot, self._num_pix_wl_tot
            ) = jifty_util.prepare_correlated_field(
                self._param_prefix,
                self._num_pix,
                cropped_border_size,
                kwargs_amplitude=self._kw_amplitude_offset,
                kwargs_fluctuations=self._kw_fluctuations,
                matern_renorm_amp=matern_renorm_amp,
                num_pix_wl=self._num_pix_wl, 
                border_wl=cropped_border_size_wl if self._num_pix_wl > 1 else 0,
                kwargs_fluctuations_wl=self._kw_fluctuations_wl,
                non_linearity='exp' if exponentiate else 'none',
                field_key=self._field_key,
                param_key_xy=self._key_xy,
                param_key_wl=self._key_wl,
                correlation_type=correlation_type,
                correlation_type_wl=correlation_type_wl,
                kernel_type=kernel_type,
            )
            self._cfm_list, self._jft_model_list = None, None  # irrelevant variables for this case
        else:
            # iterate over the bands and build a list of independent correlated fields
            self._cfm_list = []
            self._jft_model_list = []
            self._num_pix_wl_tot = self._num_pix_wl  # no cropped border in this case
            for i_wl in range(self._num_pix_wl):
                cfm, jft_model, self._num_pix_tot, _ = jifty_util.prepare_correlated_field(
                    self._param_prefix,
                    self._num_pix,
                    cropped_border_size,
                    kwargs_amplitude=self._kw_amplitude_offset,
                    kwargs_fluctuations=self._kw_fluctuations,
                    num_pix_wl=1, 
                    border_wl=0,
                    kwargs_fluctuations_wl=None,
                    non_linearity='exp' if exponentiate else 'none',
                    field_key=f'{self._field_key}_stack{i_wl}',
                    param_key_xy=self._key_xy,
                    param_key_wl=self._key_wl,
                    correlation_type=correlation_type,
                    correlation_type_wl='none',
                    kernel_type=kernel_type,
                )
                self._cfm_list.append(cfm)
                self._jft_model_list.append(jft_model)
            self._cfm, self._jft_model = None, None  # irrelevant variables for this case
        # Save the expected shapes of the field
        if self._field_type == '2d':
            self._shape_latent = (self._num_pix_tot, self._num_pix_tot)
            self._shape_direct = (self._num_pix, self._num_pix)
        elif self._field_type in ('3d', '2d_stack'):
            self._shape_latent = (self._num_pix_wl_tot, self._num_pix_tot, self._num_pix_tot)
            self._shape_direct = (self._num_pix_wl, self._num_pix, self._num_pix)
        # Nice message to say everything went smoothly
        print(f"New '{self._field_type}' CorrelatedField model "
              f"with kernel '{kernel_type}' successfully created "
              f"(final shape is {str(tuple(self._shape_direct))}).")

    def __call__(self, params):
        """Handy alias to make the class callable."""
        return self.model(params)

    def model(self, params):
        """Evaluate the model at the given parameters.
        The parameters keys are:
        - '{param_prefix}_field_xi': the field fluctuations
        - '{param_prefix}_field_zeromode': the zero mode of the field
        - '{param_prefix}_field_{dim_type}_dim_fluctuations': the field fluctuations (along the spatial dimensions)
        - '{param_prefix}_field_{dim_type}_dim_loglogavgslope': the log-log average slope of the power-spectrum (along the spatial dimensions)
        - '{param_prefix}_field_{dim_type}_dim_flexibility': the flexibility (along the spatial dimensions)
        - '{param_prefix}_field_{dim_type}_dim_asperity': the asperity (along the spatial dimensions)
        where `dim_type` is either 'xy' (spatial dimensions) or 'wl' (wavelength dimension).

        Parameters
        ----------
        params : Pytree
            Parameters values as a Pytree (e.g. dict).

        Returns
        -------
        jnp.Array
            Field model (in direct space), as 2d or 3d array.
        """
        if self._field_type in ('2d', '3d'):
            return self._model_std(self._jft_model, params)
        else:
            return self._model_stack(params)
    
    def _model_std(self, jft_model, params):
        return jft_model(params)[self._param_prefix]
    
    def _model_stack(self, params):
        model_per_band = []
        for i_wl in range(self._num_pix_wl):
            model = self._model_std(self._jft_model_list[i_wl], params)
            model_per_band.append(model)
        return jnp.stack(model_per_band, axis=0)

    def numpyro_sample_pixels(self):
        """Defines the numpyro model to be used in a Pixelated (light or mass) profile.
        
        This method is only meant to be called within a numpyro model definition,
        as it defines the prior distribution (which are all standard normally distributed) for the field parameters.

        Returns
        -------
        jnp.Array
            Field model (in direct space), as 2d array.
        """
        sample_method = getattr(self, f'_numpyro_sample_pixels_{self._kernel_type}_{self._field_type}')
        return sample_method()
        
    def _numpyro_sample_pixels_powerlaw_2d(self):
        # TODO: reduce code duplication
        # imports here to prevent the need for numpyro to be installed
        # if the CorrelatedField class is used in a non-numpyro context.
        import numpyro
        from numpyro.distributions import Normal, Independent
        # Base field parameters
        key_base = f'{self._param_prefix}_{self._field_key}'
        params = {
            f'{key_base}_xi': numpyro.sample(
                f'{key_base}_xi', 
                Independent(Normal(
                    jnp.zeros((self._num_pix_tot, self._num_pix_tot)), 
                    jnp.ones((self._num_pix_tot, self._num_pix_tot))
                ), reinterpreted_batch_ndims=2)
            ),
            f'{key_base}_zeromode': numpyro.sample(
                f'{key_base}_zeromode', 
                Normal(0., 1.),
            ),
            f'{key_base}_{self._key_xy}_fluctuations': numpyro.sample(
                f'{key_base}_{self._key_xy}_fluctuations', 
                Normal(0., 1.),
            ),
            f'{key_base}_{self._key_xy}_loglogavgslope': numpyro.sample(
                f'{key_base}_{self._key_xy}_loglogavgslope', 
                Normal(0., 1.),
            ),
        }
        # Additional optional field parameters
        if self._kw_fluctuations['flexibility'] is not None:
            params[f'{key_base}_{self._key_xy}_flexibility'] = numpyro.sample(
                f'{key_base}_{self._key_xy}_flexibility', 
                Normal(0., 1.),
            )
            if self._kw_fluctuations['asperity'] is not None:
                params[f'{key_base}_{self._key_xy}_asperity'] = numpyro.sample(
                    f'{key_base}_{self._key_xy}_asperity', 
                    Normal(0., 1.),
                )
            # we now the xy_dim is the last one in the list of correlated field grids
            # TODO: make it better to also support the wavelength dimension in a 3d or stacked 2d field
            idx_target_grid = -1
            log_vol_size = self._cfm._target_grids[idx_target_grid].harmonic_grid.log_volume.size
            params[f'{key_base}_{self._key_xy}_spectrum'] = numpyro.sample(
                f'{key_base}_{self._key_xy}_spectrum', 
                Independent(Normal(
                    jnp.zeros((log_vol_size, 2)), 
                    jnp.ones((log_vol_size, 2))
                ), reinterpreted_batch_ndims=2)
            )
        return self.model(params)
    
    def _numpyro_sample_pixels_matern_2d(self):
        # TODO: reduce code duplication
        # imports here to prevent the need for numpyro to be installed
        # if the CorrelatedField class is used in a non-numpyro context.
        import numpyro
        from numpyro.distributions import Normal, Independent
        # Base field parameters
        key_base = f'{self._param_prefix}_{self._field_key}'
        params = {
            f'{key_base}_xi': numpyro.sample(
                f'{key_base}_xi', 
                Independent(Normal(
                    jnp.zeros((self._num_pix_tot, self._num_pix_tot)), 
                    jnp.ones((self._num_pix_tot, self._num_pix_tot))
                ), reinterpreted_batch_ndims=2)
            ),
            f'{key_base}_zeromode': numpyro.sample(
                f'{key_base}_zeromode', 
                Normal(0., 1.),
            ),
            f'{key_base}_{self._key_xy}_scale': numpyro.sample(
                f'{key_base}_{self._key_xy}_scale', 
                Normal(0., 1.),
            ),
            f'{key_base}_{self._key_xy}_cutoff': numpyro.sample(
                f'{key_base}_{self._key_xy}_cutoff', 
                Normal(0., 1.),
            ),
            f'{key_base}_{self._key_xy}_loglogslope': numpyro.sample(
                f'{key_base}_{self._key_xy}_loglogslope', 
                Normal(0., 1.),
            ),
        }
        return self.model(params)
    
    def _numpyro_sample_pixels_3d(self):
        # TODO: reduce code duplication
        # imports here to prevent the need for numpyro to be installed
        # if the CorrelatedField class is used in a non-numpyro context.
        import numpyro
        from numpyro.distributions import Normal, Independent
        # Base field parameters
        key_base = f'{self._param_prefix}_{self._field_key}'
        params = {
            f'{key_base}_zeromode': numpyro.sample(
                f'{key_base}_zeromode', 
                Normal(0., 1.),
            ),
            f'{key_base}_xi': numpyro.sample(
                f'{key_base}_xi', 
                Independent(Normal(
                    jnp.zeros((self._num_pix_wl_tot, self._num_pix_tot, self._num_pix_tot)), 
                    jnp.ones((self._num_pix_wl_tot, self._num_pix_tot, self._num_pix_tot))
                ), reinterpreted_batch_ndims=3)
            ),
            f'{key_base}_{self._key_xy}_fluctuations': numpyro.sample(
                f'{key_base}_{self._key_xy}_fluctuations', 
                Normal(0., 1.),
            ),
            f'{key_base}_{self._key_xy}_loglogavgslope': numpyro.sample(
                f'{key_base}_{self._key_xy}_loglogavgslope', 
                Normal(0., 1.),
            ),
            f'{key_base}_{self._key_wl}_fluctuations': numpyro.sample(
                f'{key_base}_{self._key_wl}_fluctuations', 
                Normal(0., 1.),
            ),
            f'{key_base}_{self._key_wl}_loglogavgslope': numpyro.sample(
                f'{key_base}_{self._key_wl}_loglogavgslope', 
                Normal(0., 1.),
            ),
        }
        # Additional optional field parameters
        if self._kw_fluctuations['flexibility'] is not None:
            params[f'{key_base}_{self._key_xy}_flexibility'] = numpyro.sample(
                f'{key_base}_{self._key_xy}_flexibility', 
                Normal(0., 1.),
            )
            if self._kw_fluctuations['asperity'] is not None:
                params[f'{key_base}_{self._key_xy}_asperity'] = numpyro.sample(
                    f'{key_base}_{self._key_xy}_asperity', 
                    Normal(0., 1.),
                )
        if self._kw_fluctuations_wl['flexibility'] is not None:
            params[f'{key_base}_{self._key_wl}_flexibility'] = numpyro.sample(
                f'{key_base}_{self._key_wl}_flexibility', 
                Normal(0., 1.),
            )
            if self._kw_fluctuations_wl['asperity'] is not None:
                params[f'{key_base}_{self._key_wl}_asperity'] = numpyro.sample(
                    f'{key_base}_{self._key_wl}_asperity', 
                    Normal(0., 1.),
                )
        return self.model(params)
    
    def _numpyro_sample_pixels_2d_stack(self):
        # TODO: reduce code duplication
        # imports here to prevent the need for numpyro to be installed
        # if the CorrelatedField class is used in a non-numpyro context.
        import numpyro
        from numpyro.distributions import Normal, Independent
        # Iterate over the bands
        params = {}
        key_base = f'{self._param_prefix}_{self._field_key}'
        for i_wl in range(self._num_pix_wl):
            # Base field parameters
            params.update({
                f'{key_base}_stack{i_wl}_xi': numpyro.sample(
                    f'{key_base}_stack{i_wl}_xi', 
                    Independent(Normal(
                        jnp.zeros((self._num_pix_tot, self._num_pix_tot)), 
                        jnp.ones((self._num_pix_tot, self._num_pix_tot))
                    ), reinterpreted_batch_ndims=2)
                ),
                f'{key_base}_stack{i_wl}_zeromode': numpyro.sample(
                    f'{key_base}_stack{i_wl}_zeromode', 
                    Normal(0., 1.),
                ),
                f'{key_base}_stack{i_wl}_{self._key_xy}_fluctuations': numpyro.sample(
                    f'{key_base}_stack{i_wl}_{self._key_xy}_fluctuations', 
                    Normal(0., 1.),
                ),
                f'{key_base}_stack{i_wl}_{self._key_xy}_loglogavgslope': numpyro.sample(
                    f'{key_base}_stack{i_wl}_{self._key_xy}_loglogavgslope', 
                    Normal(0., 1.),
                ),
            })
            # Additional optional field_key parameters
            if self._kw_fluctuations['flexibility'] is not None:
                params.update(
                    {
                        f'{key_base}_stack{i_wl}_{self._key_xy}_flexibility': numpyro.sample(
                            f'{key_base}_stack{i_wl}_{self._key_xy}_flexibility', 
                            Normal(0., 1.),
                        )
                    }
                )
                if self._kw_fluctuations['asperity'] is not None:
                    params.update(
                        {
                            f'{key_base}_stack{i_wl}_{self._key_xy}_asperity': numpyro.sample(
                                f'{key_base}_stack{i_wl}_{self._key_xy}_asperity', 
                                Normal(0., 1.),
                            )
                        }
                    )
            # model = self._jft_model_list[i_wl](params)[self._param_prefix]
            # models_per_band.append(model)
        return self.model(params)
    
    def draw_realizations_from_prior(self, prng_key, num_samples=10, return_parameters=False):
        """Draw a random field realization from the prior distribution.

        Returns
        -------
        jnp.Array
            Field model (in direct space), as 2d array.
        """
        # imports here to prevent the need for numpyro to be installed
        # if the CorrelatedField class is used in a non-numpyro context.
        from numpyro.infer import Predictive
        def model():
            """This just to define a callable the 'numpyro way'"""
            self.numpyro_sample_pixels()
        # draw sample from the latent space (i.e. standard normal samples)
        prior_samples = Predictive(model, num_samples=num_samples)(prng_key)
        # evaluate the model at the prior samples
        model_samples = jax.vmap(self)(prior_samples)
        # if there is only one sample, remove the first dimension
        if num_samples == 1:
            model_samples = jnp.squeeze(model_samples, axis=0)
        if return_parameters:
            return model_samples, prior_samples
        return model_samples
    
    @property
    def correlated_field_maker(self):
        return self._cfm
    
    @property
    def num_pix_field(self):
        return self._num_pix_tot
    
    @property
    def num_pix(self):
        return self._num_pix
    
    @property
    def num_wl_field(self):
        return self._num_pix_wl_tot
    
    @property
    def num_wl(self):
        return self._num_pix_wl

    @property
    def final_shape(self):
        return self._shape_direct
    
    @property
    def field_latent(self):
        return self._shape_latent
    


# Defines a gaussian profile
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LightModel.Profiles module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp

from herculens.Util import param_util


__all__ = ['Gaussian', 'GaussianEllipse']


class Gaussian(object):
    """
    class for Gaussian light profile
    The two-dimensional Gaussian profile amplitude is defined such that the 2D integral leads to the 'amp' value.

    profile name in LightModel module: 'GAUSSIAN'
    """
    param_names = ['amp', 'sigma', 'center_x', 'center_y']
    lower_limit_default = {'amp': 0, 'sigma': 0, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'amp': 1000, 'sigma': 100, 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}

    def function(self, x, y, amp, sigma, center_x=0, center_y=0):
        """
        surface brightness per angular unit

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness at (x, y)
        """
        c = amp / (2 * np.pi * sigma**2)
        R2 = (x - center_x) ** 2 / sigma**2 + (y - center_y) ** 2 / sigma**2
        return c * jnp.exp(-R2 / 2.)

    def total_flux(self, amp, sigma, center_x=0, center_y=0):
        """
        integrated flux of the profile

        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param center_x: center of profile
        :param center_y: center of profile
        :return: total flux
        """
        return amp

    def light_3d(self, r, amp, sigma):
        """
        3D brightness per angular volume element

        :param r: 3d distance from center of profile
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :return: 3D brightness per angular volume element
        """
        amp3d = amp / np.sqrt(2 * sigma**2) / np.sqrt(np.pi)
        sigma3d = sigma
        return self.function(r, 0, amp3d, sigma3d)
    
    @property
    def num_amplitudes(self):
        return 1


class GaussianEllipse(object):
    """
    class for Gaussian light profile with ellipticity

    profile name in LightModel module: 'GAUSSIAN_ELLIPSE'
    """
    param_names = ['amp', 'sigma', 'e1', 'e2', 'center_x', 'center_y']
    lower_limit_default = {'amp': 0, 'sigma': 0, 'e1': -0.5, 'e2': -0.5, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'amp': 1000, 'sigma': 100, 'e1': -0.5, 'e2': -0.5, 'center_x': 100, 'center_y': 100}

    def __init__(self):
        self.gaussian = Gaussian()

    def function(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness at (x, y)
        """
        x_, y_ = param_util.transform_e1e2_product_average(x, y, e1, e2, center_x, center_y)
        return self.gaussian.function(x_, y_, amp, sigma, center_x=0, center_y=0)

    def total_flux(self, amp, sigma=None, e1=None, e2=None, center_x=None, center_y=None):
        """
        total integrated flux of profile

        :param x: coordinate on the sky
        :param y: coordinate on the sky
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :param center_x: center of profile
        :param center_y: center of profile
        :return: total flux
        """
        return self.gaussian.total_flux(amp, sigma, center_x, center_y)

    def light_3d(self, r, amp, sigma, e1=0, e2=0):
        """
        3D brightness per angular volume element

        :param r: 3d distance from center of profile
        :param amp: amplitude, such that 2D integral leads to this value
        :param sigma: sigma of Gaussian in each direction
        :param e1: eccentricity modulus
        :param e2: eccentricity modulus
        :return: 3D brightness per angular volume element
        """
        return self.gaussian.light_3d(r, amp, sigma=sigma)

    @property
    def num_amplitudes(self):
        return 1
    


# Defines en dual pseudo isothermal elliptical (dPIE) mass profile
# 
# Copyright (c) 2024, herculens developers and contributors

__author__ = 'aymgal'


import numpy as np
import jax.numpy as jnp
from functools import partial
from jax import jit
from jax.tree_util import register_pytree_node


__all__ = [
    'ZeroMass', 
]


class ZeroMass(object):
    """
    Dummy profile that does not have any mass associated with it.
    """
    param_names = []
    lower_limit_default = {}
    upper_limit_default = {}
    fixed_default = {}
    
    @partial(jit, static_argnums=(0,))
    def function(self, x, y, **kwargs):
        return jnp.zeros_like(x)

    @partial(jit, static_argnums=(0,))
    def derivatives(self, x, y, **kwargs):
        return jnp.zeros_like(x), jnp.zeros_like(y)
    
    @partial(jit, static_argnums=(0,))
    def hessian(self, x, y, **kwargs):
        return jnp.zeros_like(x), jnp.zeros_like(y), jnp.zeros_like(y)
    
def flatten_func(self):
    children = ()
    aux_data = {}
    return (children, aux_data)

def unflatten_func(aux_data, children):
    return ZeroMass()

register_pytree_node(ZeroMass, flatten_func, unflatten_func)



# Handles different convolution methods
#
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the ImSim.Numerics module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp
import jax.scipy as jsp

from utax.convolution.classes import GaussianFilter
from utax.convolution.functions import build_convolution_matrix

from herculens.Util import util, kernel_util, image_util


__all__ = ['PixelKernelConvolution', 'SubgridKernelConvolution', 'GaussianConvolution']


class PixelKernelConvolution(object):
    """
    class to compute convolutions for a given pixelized kernel
    """

    _conv_types = ['jax_scipy_fft', 'jax_scipy', 'matrix']

    def __init__(self, kernel, convolution_type='jax_scipy_fft', output_shape=None):
        """

        :param kernel: 2d array, convolution kernel
        """
        self._kernel = kernel
        if convolution_type not in self._conv_types:
            raise ValueError(f"Convolution type '{convolution_type}' not supported "
                             f"(should in {self._conv_types}).")
        self._conv_type = convolution_type
        if self._conv_type == 'matrix':
            if output_shape is None:
                raise ValueError("An output shape must be provided to build the convolution matrix.")
            self._conv_matrix  = build_convolution_matrix(kernel, output_shape)
            self._output_shape = output_shape
        else:
            self._conv_matrix  = None
            self._output_shape = None

    def pixel_kernel(self, num_pix=None):
        """
        access pixelated kernel

        :param num_pix: size of returned kernel (odd number per axis). If None, return the original kernel.
        :return: pixel kernel centered
        """
        if num_pix is not None:
            return kernel_util.cut_psf(self._kernel, num_pix)
        return self._kernel

    def convolution2d(self, image):
        """

        :param image: 2d array (image) to be convolved
        :return: fft convolution
        """
        if self._conv_type == 'jax_scipy_fft':
            return jsp.signal.fftconvolve(image, self._kernel, mode='same')
        elif self._conv_type == 'jax_scipy':
            return jsp.signal.convolve2d(image, self._kernel, mode='same')
        elif self._conv_type == 'matrix':
            return self._conv_matrix.dot(image.flatten()).reshape(*self._output_shape)

    def re_size_convolve(self, image_low_res, image_high_res=None):
        """

        :param image_high_res: supersampled image/model to be convolved on a regular pixel grid
        :return: convolved and re-sized image
        """
        return self.convolution2d(image_low_res)

    @property
    def convolution_matrix(self):
        return self._conv_matrix



class SubgridKernelConvolution(object):
    """
    class to compute the convolution on a supersampled grid with partial convolution computed on the regular grid
    """
    def __init__(self, kernel_supersampled, supersampling_factor, supersampling_kernel_size=None):
        """

        :param kernel_supersampled: kernel in supersampled pixels
        :param supersampling_factor: supersampling factor relative to the image pixel grid
        :param supersampling_kernel_size: number of pixels (in units of the image pixels) that are convolved with the
        supersampled kernel
        """
        n_high = len(kernel_supersampled)
        self._supersampling_factor = supersampling_factor
        numPix = int(n_high / self._supersampling_factor)
        if supersampling_kernel_size is None:
            kernel_low_res, kernel_high_res = np.zeros((3, 3)), kernel_supersampled
            self._low_res_convolution = False
        else:
            kernel_low_res, kernel_high_res = kernel_util.split_kernel(kernel_supersampled, supersampling_kernel_size,
                                                                       self._supersampling_factor)
            self._low_res_convolution = True
        self._low_res_conv = PixelKernelConvolution(kernel_low_res)
        self._high_res_conv = PixelKernelConvolution(kernel_high_res)

    def convolution2d(self, image):
        """

        :param image: 2d array (high resoluton image) to be convolved and re-sized
        :return: convolved image
        """

        image_high_res_conv = self._high_res_conv.convolution2d(image)
        image_resized_conv = image_util.re_size(image_high_res_conv, self._supersampling_factor)
        if self._low_res_convolution is True:
            image_resized = image_util.re_size(image, self._supersampling_factor)
            image_resized_conv += self._low_res_conv.convolution2d(image_resized)
        return image_resized_conv

    def re_size_convolve(self, image_low_res, image_high_res):
        """

        :param image_high_res: supersampled image/model to be convolved on a regular pixel grid
        :return: convolved and re-sized image
        """
        image_high_res_conv = self._high_res_conv.convolution2d(image_high_res)
        image_resized_conv = image_util.re_size(image_high_res_conv, self._supersampling_factor)
        if self._low_res_convolution is True:
            image_resized_conv += self._low_res_conv.convolution2d(image_low_res)
        return image_resized_conv


class GaussianConvolution(object):
    """
    class to perform a convolution a 2d Gaussian
    """

    def __init__(self, sigma, pixel_scale, supersampling_factor=1,
                 supersampling_convolution=False, truncation=2):
        self._sigma = sigma / pixel_scale
        if supersampling_convolution is True:
            self._sigma *= supersampling_factor
        self._truncation = truncation
        self._pixel_scale = pixel_scale
        self._supersampling_factor = supersampling_factor
        self._supersampling_convolution = supersampling_convolution
        self._gaussian_filter = GaussianFilter(self._sigma, self._truncation)

    def convolution2d(self, image):
        """
        2d convolution

        :param image: 2d numpy array, image to be convolved
        :return: convolved image, 2d numpy array
        """
        return self._gaussian_filter(image)

    def re_size_convolve(self, image_low_res, image_high_res):
        """

        :param image_high_res: supersampled image/model to be convolved on a regular pixel grid
        :return: convolved and re-sized image
        """
        if self._supersampling_convolution:
            image_high_res_conv = self.convolution2d(image_high_res)
            image_resized_conv = image_util.re_size(image_high_res_conv, self._supersampling_factor)
        else:
            image_resized_conv = self.convolution2d(image_low_res)
        return image_resized_conv

    def pixel_kernel(self, num_pix):
        """
        computes a pixelized kernel from the MGE parameters

        :param num_pix: int, size of kernel (odd number per axis)
        :return: pixel kernel centered
        """
        x, y = util.make_grid(numPix=num_pix, deltapix=self._pixel_scale)
        sigma = self._sigma
        diff_square = (x - center_x) ** 2 / sigma**2 + (y - center_y) ** 2 / sigma**2
        kernel = jnp.exp(- diff_square / 2.)
        kernel = util.array2image(kernel)
        return kernel / jnp.sum(kernel)



# Defines regularization choices
# 
# Copyright (c) 2023, herculens developers and contributors


__author__ = 'aymgal'


import jax.numpy as jnp

from herculens.RegulModel.Methods.base import BaseRegularization


__all__ = [
    'Positivity',
    'Negativity',
]


class Positivity(BaseRegularization):

    param_names = ['strength']
    lower_limit_default = {'strength': 0}
    upper_limit_default = {'strength': 1e8}
    fixed_default = {key: True for key in param_names}

    def __init__(self, model_type, profile_index, mass_form=None):
        super().__init__(model_type, profile_index, mass_form=mass_form)

    def initialize(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass

    def log_prob(self, kwargs_params, strength=0):
        pixels = self.get_pixel_params(kwargs_params)
        log_prob = - strength * jnp.abs(jnp.sum(jnp.minimum(0., pixels)))
        return log_prob


class Negativity(BaseRegularization):

    param_names = ['strength']
    lower_limit_default = {'strength': 0}
    upper_limit_default = {'strength': 1e8}
    fixed_default = {key: True for key in param_names}

    def __init__(self, model_type, profile_index, mass_form=None):
        super().__init__(model_type, profile_index, mass_form=mass_form)

    def initialize(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass

    def log_prob(self, kwargs_params, strength=0):
        pixels = self.get_pixel_params(kwargs_params)
        log_prob = - strength * jnp.abs(jnp.sum(jnp.maximum(0., pixels)))
        return log_prob



# Defines the model of a strong lens
# 
# Copyright (c) 2022, herculens developers and contributors
# based on the ImSim module from lenstronomy (version 1.9.3)

__author__ = 'aymgal'


__all__ = ['ProbabilisticModel']


class BaseProbModel(object):
    """Base class for probabilistic model"""

    def model(self):
        raise NotImplementedError("Must be implemented by user class")
        
    def params2kwargs(self, params):
        raise NotImplementedError("`params2kwargs` method must be implemented.")

    def log_prob(self, params):
        raise NotImplementedError("`log_prob` method must be implemented.")
    
    def sample_prior(self, num_samples, seed=0):
        raise NotImplementedError("`draw_samples` method must be implemented.")



# Class to plot a lens model
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal'


import copy
import warnings
import numpy as np
import jax.numpy as jnp
import matplotlib.pyplot as plt
from scipy import ndimage
import matplotlib as mpl
from matplotlib.colors import Normalize, LogNorm, TwoSlopeNorm
from matplotlib.patches import FancyArrowPatch, ArrowStyle

from herculens.LensImage.lens_image import LensImage, LensImage3D
from herculens.LensImage.lens_image_multiplane import MPLensImage
from herculens.Util.plot_util import nice_colorbar, nice_colorbar_residuals, contour_with_legend
from herculens.Util import model_util


# Some general default for plotting
plt.rc('image', interpolation='none', origin='lower')  # for imshow


__all__ = ['Plotter']


class Plotter(object):
    """
    Helper class to plot the results of a LensImage model.

    Parameters
    ----------
    data_name : str, optional
        The name of the data, by default None.
    base_fontsize : int, optional
        The base fontsize for the plot, by default 14.
    flux_log_scale : bool, optional
        Whether to use a logarithmic scale for the flux, by default True.
    flux_vmin : float, optional
        The minimum value for the flux, by default None.
    flux_vmax : float, optional
        The maximum value for the flux, by default None.
    res_vmax : int, optional
        The maximum value for the residual, by default 6.
    cmap_flux : str, optional
        The colormap for the flux, by default None.
    """

    # Define some custom colormaps
    cmap_flux = copy.copy(plt.get_cmap('magma'))
    cmap_flux.set_under('black')
    cmap_flux.set_over('white')
    cmap_flux.set_bad('black')
    cmap_flux_alt = copy.copy(cmap_flux)
    cmap_flux_alt.set_bad('#222222')  # to emphasize non-positive pixels in log scale
    cmap_res = plt.get_cmap('RdBu_r')
    cmap_corr = plt.get_cmap('RdYlGn')
    cmap_default = plt.get_cmap('viridis')
    cmap_deriv1 = plt.get_cmap('cividis')
    cmap_deriv2 = plt.get_cmap('inferno')
    cmap_bw = copy.copy(plt.get_cmap('gray'))
    cmap_bw.set_under('black')
    cmap_bw.set_over('white')
    cmap_bw.set_bad('black')

    def __init__(self, data_name=None, base_fontsize=14, flux_log_scale=True, 
                 flux_vmin=None, flux_vmax=None, res_vmax=6, cmap_flux=None,
                 ref_lens_image=None, ref_kwargs_result=None):
        self.data_name = data_name
        self.base_fontsize = base_fontsize
        self.flux_log_scale = flux_log_scale
        if self.flux_log_scale is True:
            self.norm_flux = LogNorm(flux_vmin, flux_vmax)
        else:
            self.norm_flux = None
        self.norm_res = Normalize(-res_vmax, res_vmax)
        self.norm_corr = TwoSlopeNorm(0)
        if cmap_flux is not None:
            self.cmap_flux = cmap_flux
            self.cmap_flux_alt = cmap_flux
        if ref_lens_image is not None and ref_kwargs_result is None:
            raise ValueError("If a reference lens image is provided, "
                             "the reference kwargs_result must also be provided.")
        self.ref_lens_image = ref_lens_image
        self.ref_kwargs_result = ref_kwargs_result

    def set_data(self, data):
        self._data = data

    def set_ref_source(self, ref_source, plt_extent=None):
        if self.ref_lens_image is not None:
            raise ValueError("Reference source already set from a LensImage instance.")
        self._ref_source = ref_source
        self._ref_source_extent = plt_extent

    def set_ref_lens_light(self, ref_lens_light):
        self._ref_lens_light = ref_lens_light

    def set_ref_pixelated_potential(self, ref_potential):
        self._ref_pixel_pot = ref_potential

    def plot_flux(self, image, title=None):
        fig, ax = plt.subplots(1, 1)
        im = ax.imshow(image, extent=None, cmap=self.cmap_flux, norm=self.norm_flux)
        if title is not None:
            ax.set_title(title, fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                      colorbar_kwargs={'orientation': 'horizontal'})
        return fig

    def model_summary(
            self, lens_image, kwargs_result,
            show_image=True, show_source=True, 
            show_lens_light=False, show_lens_potential=False, show_lens_others=False,
            only_pixelated_potential=False, shift_pixelated_potential='none',
            likelihood_mask=None, potential_mask=None, apply_mask_to_model=False,
            show_lens_lines=False, supersampling_lens_lines=3,
            show_shear_field=False, show_lens_position=False,
            kwargs_grid_source=None,
            lock_colorbars=False, masked_residuals=True,
            vmin_pot=None, vmax_pot=None,
            k_source=None, k_lens=None, k_lens_light=None, k_point_source=None,  # for single-plane models
            k_light=None, k_mass=None, k_planes=None,  # for multi-plane models
            kwargs_noise=None, 
            figsize=None, show_plot=True,

            # TODO: this is a dirty quick fix to correctly read multi-band heterogenous LensImage models
            # This will not be necessary once the LensImage3D class is properly supported.
            adapted_source_pixels_coords=None,
        ):
        """
        Generate a summary plot of the lens model.

        Parameters
        ----------
        lens_image : LensImage
            A herculens.LensImage instance.
        kwargs_result : dict
            Nested dictionary containing all model parameters.
        show_image : bool, optional
            Whether to show the lens image, by default True.
        show_source : bool, optional
            Whether to show the source model, by default True.
        show_lens_light : bool, optional
            Whether to show the lens light model, by default False.
        show_lens_potential : bool, optional
            Whether to show the lens potential model, by default False.
        show_lens_others : bool, optional
            Whether to show other lens models, by default False.
        only_pixelated_potential : bool, optional
            Whether to show only the pixelated potential, by default False.
        shift_pixelated_potential : str, optional
            The type of shift to apply to the pixelated potential, by default 'none'.
        likelihood_mask : ndarray, optional
            The likelihood mask, by default None.
        potential_mask : ndarray, optional
            The potential mask, by default None.
        apply_mask_to_model : bool, optional
            Whether to apply the likelihood mask to the model image, in order
            to hide away regions of the model that were not used in the likelihood.
            By default False.
        show_lens_lines : bool, optional
            Whether to show lens lines, by default False.
        show_shear_field : bool, optional
            Whether to show the shear field, by default False.
        show_lens_position : bool, optional
            Whether to show the lens position, by default False.
        kwargs_grid_source : dict, optional
            The grid source parameters, by default None.
        lock_colorbars : bool, optional
            Whether to lock the colorbars, by default False.
        masked_residuals : bool, optional
            Whether to show masked residuals, by default True.
        vmin_pot : float, optional
            The minimum potential value, by default None.
        vmax_pot : float, optional
            The maximum potential value, by default None.
        k_lens : float, optional
            The lens model normalization factor, by default None.
        kwargs_noise : dict, optional
            Parameters given to Noise.C_D_model(mode, **kwargs_noise), by default None.
        show_plot : bool, optional
            Whether to call plt.show(), by default True.

        Returns
        -------
        type
            The summary plot.
        """
        if isinstance(lens_image, MPLensImage) and any([
            show_source, show_lens_light, show_lens_potential, show_lens_others,
            show_shear_field, show_lens_position, show_lens_lines,
        ]):
            raise NotImplementedError("The full plotting of multi-plane lens models, "
                                      "other than just the data model (observed image plane), "
                                      "is not supported yet.")
            
        n_cols = 3
        n_rows = sum([show_image, show_source, show_lens_light, 
                      show_lens_potential, show_lens_others])
        
        # extent = lens_image.Grid.extent
        extent = lens_image.Grid.plt_extent

        ##### PREPARE IMAGES #####

        if kwargs_noise is None:
            kwargs_noise = {}
            
        if show_image:
            # create the resulting model image
            if isinstance(lens_image, (LensImage, LensImage3D)):
                model = lens_image.model(
                    **kwargs_result, 
                    k_lens=k_lens,
                    k_lens_light=k_lens_light,
                    k_source=k_source,
                    k_point_source=k_point_source,
                    adapted_source_pixels_coords=adapted_source_pixels_coords,
                )
            elif isinstance(lens_image, MPLensImage):
                model = lens_image.model(
                    **kwargs_result, 
                    k_mass=k_mass,
                    k_light=k_light,
                    k_planes=k_planes,
                )
            if likelihood_mask is None:
                mask_bool = False
                likelihood_mask = np.ones_like(model)
            else:
                mask_bool = True
            # create a mask with NaNs such that unmasked areasa appear transparent 
            likelihood_mask_nans = np.nan*np.copy(likelihood_mask)
            likelihood_mask_nans[likelihood_mask == 0] = 0

            if hasattr(self, '_data'):
                data = self._data
            else:
                data = np.zeros_like(model)

        if show_source:
            kwargs_source = copy.deepcopy(kwargs_result['kwargs_source'])
            if lens_image.SourceModel.has_pixels or kwargs_grid_source is not None:
                if kwargs_grid_source is not None:
                    grid_src = lens_image.Grid.create_model_grid(**kwargs_grid_source)
                    x_grid_src, y_grid_src = grid_src.pixel_coordinates
                    src_extent = grid_src.plt_extent
                else:
                    x_grid_src, y_grid_src, src_extent = lens_image.get_source_coordinates(
                        kwargs_result['kwargs_lens'], k_lens=k_lens, return_plt_extent=True
                    )
                source_model = lens_image.eval_source_surface_brightness(
                    x_grid_src, y_grid_src, 
                    kwargs_source, kwargs_lens=kwargs_result['kwargs_lens'],
                    k=k_source, k_lens=k_lens, de_lensed=True,
                    adapted_pixels_coords=adapted_source_pixels_coords,
                ) * lens_image.Grid.pixel_area
            else:
                source_model = lens_image.source_surface_brightness(
                    kwargs_source, kwargs_lens=kwargs_result['kwargs_lens'], 
                    de_lensed=True, unconvolved=True, 
                    k=k_source, k_lens=k_lens,
                )
                x_grid_src, y_grid_src = lens_image.ImageNumerics.coordinates_evaluate
                src_extent = extent

            if self.ref_lens_image is not None:
                ref_source = self.ref_lens_image.eval_source_surface_brightness(
                    x_grid_src, y_grid_src, 
                    self.ref_kwargs_result['kwargs_source'], 
                    de_lensed=True, 
                    adapted_pixels_coords=adapted_source_pixels_coords,
                ) * lens_image.Grid.pixel_area
                ref_src_extent = src_extent
                show_source_diff = True
            elif hasattr(self, '_ref_source'):
                ref_source = self._ref_source
                if source_model.shape != ref_source.shape:
                    warnings.warn("Reference source does not have the same shape as model source.")
                    show_source_diff = False
                else:
                    show_source_diff = True
                ref_src_extent = self._ref_source_extent
            else:
                ref_source = None
                ref_src_extent = None
                show_source_diff = False

            if len(lens_image.PointSourceModel.type_list) > 0:
                #TODO: support several point source models
                ps0_params = kwargs_result['kwargs_point_source'][0]
                all_ps_src_x, all_ps_src_y = lens_image.PointSourceModel.get_source_plane_points(
                    kwargs_result['kwargs_point_source'],
                    kwargs_lens=kwargs_result['kwargs_lens'],
                    with_amplitude=False,
                )
                ps_src_pos = (all_ps_src_x[0], all_ps_src_y[0])
            else:
                ps_src_pos = None

        if show_lens_light:
            kwargs_lens_light = copy.deepcopy(kwargs_result['kwargs_lens_light'])
            if lens_image.LensLightModel.has_pixels:
                ll_idx = lens_image.LensLightModel.pixelated_index
                lens_light_model = kwargs_lens_light[ll_idx]['pixels']
            else:
                lens_light_model = lens_image.lens_surface_brightness(kwargs_lens_light, unconvolved=True)
            
            if hasattr(self, '_ref_lens_light'):
                ref_lens_light = self._ref_lens_light
                if lens_light_model.shape != ref_lens_light.shape:
                    warnings.warn("Reference lens light does not have the same shape as model lens light.")
                    show_lens_light_diff = False
                else:
                    show_lens_light_diff = True
            else:
                ref_lens_light = None
                show_lens_light_diff = False

        if show_lens_potential or show_lens_others:
            kwargs_lens = copy.deepcopy(kwargs_result['kwargs_lens'])
            pot_idx = lens_image.MassModel.pixelated_index if only_pixelated_potential else None
            if pot_idx is not None and only_pixelated_potential:
                x_grid_lens, y_grid_lens = lens_image.MassModel.pixel_grid.pixel_coordinates
                potential_model = kwargs_lens[pot_idx]['pixels']
            else:
                x_grid_lens, y_grid_lens = lens_image.Grid.pixel_coordinates
                if show_lens_potential:
                    potential_model = lens_image.MassModel.potential(x_grid_lens, y_grid_lens, 
                                                                     kwargs_lens, k=pot_idx)
            alpha_x, alpha_y = lens_image.MassModel.alpha(x_grid_lens, y_grid_lens, 
                                                          kwargs_lens, k=pot_idx)
            kappa = lens_image.MassModel.kappa(x_grid_lens, y_grid_lens, 
                                               kwargs_lens, k=pot_idx)
            #kappa = ndimage.gaussian_filter(kappa, 1)
            magnification = lens_image.MassModel.magnification(x_grid_lens, y_grid_lens, kwargs_lens)
            
            if potential_mask is None:
                potential_mask = np.ones_like(x_grid_lens)

            # here we know that there are no perturbations in the reference potential
            if hasattr(self, '_ref_pixel_pot') and show_lens_potential:
                ref_potential = self._ref_pixel_pot
                if ref_potential.shape != potential_model.shape:
                    warnings.warn("Reference potential does not have the same shape as model potential.")
                    show_pot_diff = False
                else:
                    show_pot_diff = True
            
                if shift_pixelated_potential == 'min':
                    min_in_mask = potential_model[potential_mask == 1].min()
                    potential_model = potential_model - min_in_mask
                    ref_min_in_mask = ref_potential[potential_mask == 1].min()
                    ref_potential = ref_potential - ref_min_in_mask
                    print("delta_psi shift by min:", min_in_mask)
                if shift_pixelated_potential == 'max':
                    max_in_mask = potential_model[potential_mask == 1].max()
                    potential_model = potential_model - max_in_mask
                    ref_max_in_mask = ref_potential[potential_mask == 1].max()
                    ref_potential = ref_potential - ref_max_in_mask
                    print("delta_psi shift by max:", max_in_mask)
                elif shift_pixelated_potential == 'mean':
                    mean_in_mask = potential_model[potential_mask == 1].mean()
                    potential_model = potential_model - mean_in_mask
                    ref_mean_in_mask = ref_potential[potential_mask == 1].mean()
                    ref_potential = ref_potential - ref_mean_in_mask
                    print("delta_psi shift by mean values:", mean_in_mask, ref_mean_in_mask)

            else:
                ref_potential = None
                show_pot_diff = False

        if show_lens_lines:
            clines, caustics, centers = model_util.critical_lines_caustics(
                lens_image, 
                kwargs_result['kwargs_lens'], 
                return_lens_centers=True,
                supersampling=supersampling_lens_lines,
            )


        ##### BUILD UP THE PLOTS #####
        if figsize is None:
            figsize = (15, n_rows*5)
        fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)
        if len(axes.shape) == 1:
            axes = axes[None, :] # add first axis so axes is always 2d array
        i_row = 0

        if show_image:
            norm_flux = self._get_norm_for_model(model, lock_colorbars)

            ##### IMAGING DATA AND MODEL IMAGE #####
            ax = axes[i_row, 0]
            im = ax.imshow(data, extent=extent, cmap=self.cmap_flux, norm=norm_flux)
            im.set_rasterized(True)
            if mask_bool is True:
                ax.contour(likelihood_mask, extent=extent, levels=[0], 
                           colors='white', alpha=0.3, linewidths=0.5)
            if show_lens_position and 'kwargs_lens_light' in kwargs_result:
                plotted_centers = []
                for kw in kwargs_result['kwargs_lens_light']:
                    if 'center_x' in kw:
                        center = (kw['center_x'], kw['center_y'])
                        if center not in plotted_centers:
                            ax.plot(*center, linestyle='none', color='black', 
                                    markeredgecolor='black', markersize=15, marker='o', 
                                    fillstyle='none', markeredgewidth=1)
                        plotted_centers.append(center)
            if show_lens_lines:
                for curve in clines:
                    ax.plot(curve[0], curve[1], linewidth=0.8, color='white')
                ax.scatter(*centers, s=20, c='gray', marker='+', linewidths=0.5)
            if show_shear_field:
                shear_field = model_util.shear_deflection_field(lens_image, kwargs_result['kwargs_lens'], num_pixels=8)
                if shear_field is not None:
                    x_field, y_field, gx_field, gy_field = shear_field
                    ax.quiver(
                        x_field, y_field, # + overall_shift, 
                        gx_field, gy_field, 
                        scale=0.2, 
                        width=0.05,
                        scale_units='xy', units='xy',
                        pivot='middle',
                        headaxislength=0, headlength=0,
                        color='white', alpha=0.3,
                    )
                    ax.set_xlim(extent[0], extent[1])
                    ax.set_ylim(extent[2], extent[3])
                else:
                    print("Warning: no external shear to plot have been found.")
            data_title = self.data_name if self.data_name is not None else "data"
            ax.set_title(data_title, fontsize=self.base_fontsize)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})
            
            ax = axes[i_row, 1]
            if apply_mask_to_model is True:
                model_to_show = model * likelihood_mask
            else:
                model_to_show = model
            im = ax.imshow(model_to_show, extent=extent, cmap=self.cmap_flux, norm=norm_flux)
            im.set_rasterized(True)
            ax.set_title("model", fontsize=self.base_fontsize)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})

            ax = axes[i_row, 2]
            model_residuals, residuals = lens_image.normalized_residuals(
                data, model, kwargs_noise=kwargs_noise, mask=likelihood_mask,
            )
            if masked_residuals is True:
                residuals_plot = model_residuals
            else:
                residuals_plot = residuals
            red_chi2 = lens_image.reduced_chi2(
                data, model, kwargs_noise=kwargs_noise, mask=likelihood_mask,
            )
            im = ax.imshow(residuals_plot, cmap=self.cmap_res, extent=extent, norm=self.norm_res)
            im.set_rasterized(True)
            if mask_bool is True and masked_residuals is False:
                ax.contour(likelihood_mask, extent=extent, levels=[0], 
                           colors='black', alpha=0.5, linewidths=0.5)
            ax.set_title(r"(f${}_{\rm data}$ - f${}_{\rm model})/\sigma$", fontsize=self.base_fontsize)
            nice_colorbar_residuals(im, residuals_plot, position='top', pad=0.4, size=0.2, 
                                    vmin=self.norm_res.vmin, vmax=self.norm_res.vmax,
                                    colorbar_kwargs={'orientation': 'horizontal'})
            text = r"$\chi^2_\nu={:.2f}$".format(red_chi2)
            ax.text(0.05, 0.05, text, color='black', fontsize=self.base_fontsize-4, 
                    horizontalalignment='left', verticalalignment='bottom',
                    transform=ax.transAxes, bbox={'color': 'white', 'alpha': 0.8})
            i_row += 1

        if show_source:
            norm_flux = self._get_norm_for_model(source_model, lock_colorbars)

            ##### UNLENSED AND UNCONVOLVED SOURCE MODEL #####
            ax = axes[i_row, 0]
            if ref_source is not None:
                im = ax.imshow(ref_source, extent=ref_src_extent, cmap=self.cmap_flux_alt, norm=norm_flux) #, vmax=vmax)
                im.set_rasterized(True)
                ax.set_title("ref. source", fontsize=self.base_fontsize)
                nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                              colorbar_kwargs={'orientation': 'horizontal'})
            else:
                ax.axis('off')
            ax = axes[i_row, 1]
            im = ax.imshow(source_model, extent=src_extent, cmap=self.cmap_flux_alt, norm=norm_flux) #, vmax=vmax)
            #im = ax.imshow(source_model, extent=extent, cmap=self.cmap_flux_alt, norm=LogNorm(1e-5))
            im.set_rasterized(True)
            ax.set_title("source model", fontsize=self.base_fontsize)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})
            if show_lens_lines:
                for curve in caustics:
                    ax.plot(curve[0], curve[1], linewidth=0.8, color='white')
                    # force x, y limits to stay the same as 
                    ax.set_xlim(src_extent[0], src_extent[1])
                    ax.set_ylim(src_extent[2], src_extent[3])
            if ps_src_pos is not None:
                ax.scatter(*ps_src_pos, s=100, c='tab:green', marker='*', linewidths=0.5, 
                           label="point source")
                ax.legend()
            ax = axes[i_row, 2]
            if ref_source is not None and show_source_diff is True:
                diff = source_model - ref_source
                vmax_diff = ref_source.max() / 10.
                im = ax.imshow(diff, extent=src_extent, 
                               cmap=self.cmap_res, norm=Normalize(-vmax_diff, vmax_diff))
                im.set_rasterized(True)
                ax.set_title(r"s${}_{\rm model}$ - s${}_{\rm ref}$", fontsize=self.base_fontsize)
                nice_colorbar_residuals(im, diff, position='top', pad=0.4, size=0.2, 
                                        vmin=-vmax_diff, vmax=vmax_diff,
                                        colorbar_kwargs={'orientation': 'horizontal'})
            else:
                ax.axis('off')
            i_row += 1

        if show_lens_light:
            norm_flux = self._get_norm_for_model(lens_light_model, lock_colorbars)

            ##### UNLENSED AND UNCONVOLVED SOURCE MODEL #####
            ax = axes[i_row, 0]
            if ref_lens_light is not None:
                im = ax.imshow(ref_lens_light, extent=extent, cmap=self.cmap_flux_alt, norm=norm_flux) #, vmax=vmax)
                im.set_rasterized(True)
                ax.set_title("ref. lens light", fontsize=self.base_fontsize)
                nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                              colorbar_kwargs={'orientation': 'horizontal'})
            else:
                ax.axis('off')
            ax = axes[i_row, 1]
            im = ax.imshow(lens_light_model, extent=extent, cmap=self.cmap_flux_alt, norm=norm_flux) #, vmax=vmax)
            #im = ax.imshow(lens_light_model, extent=extent, cmap=self.cmap_flux_alt, norm=LogNorm(1e-5))
            im.set_rasterized(True)
            ax.set_title("lens light model", fontsize=self.base_fontsize)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})
            ax = axes[i_row, 2]
            if ref_lens_light is not None:
                diff = lens_light_model - ref_lens_light
                vmax_diff = ref_lens_light.max() / 10.
                im = ax.imshow(diff, extent=extent, 
                               cmap=self.cmap_res, norm=Normalize(-vmax_diff, vmax_diff))
                im.set_rasterized(True)
                ax.set_title(r"l${}_{\rm model}$ - l${}_{\rm ref}$", fontsize=self.base_fontsize)
                nice_colorbar_residuals(im, diff, position='top', pad=0.4, size=0.2, 
                                        vmin=-vmax_diff, vmax=vmax_diff,
                                        colorbar_kwargs={'orientation': 'horizontal'})
            else:
                ax.axis('off')
            i_row += 1

        if show_lens_potential:

            ##### PIXELATED POTENTIAL PERTURBATIONS #####
            ax = axes[i_row, 0]
            if ref_potential is not None:
                im = ax.imshow(ref_potential * potential_mask, extent=extent,
                               vmin=vmin_pot, vmax=vmax_pot,
                               cmap=self.cmap_default)
                im.set_rasterized(True)
                ax.set_title(r"$\psi_{\rm pix, ref}$", fontsize=self.base_fontsize)
                nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                              colorbar_kwargs={'orientation': 'horizontal'})
                ax.imshow(likelihood_mask_nans, extent=extent, cmap='gray_r', vmin=0, vmax=1)
            else:
                ax.axis('off')

            ax = axes[i_row, 1]
            if apply_mask_to_model is True:
                potential_model_to_show = model * potential_mask
            else:
                potential_model_to_show = potential_model
            im = ax.imshow(potential_model_to_show, extent=extent,
                           vmin=vmin_pot, vmax=vmax_pot,
                           cmap=self.cmap_default)
            ax.set_title(r"$\psi_{\rm pix}$", fontsize=self.base_fontsize)
            im.set_rasterized(True)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})
            ax.imshow(likelihood_mask_nans, extent=extent, cmap='gray_r', vmin=0, vmax=1)

            ax = axes[i_row, 2]
            if ref_potential is not None and show_pot_diff is True:
                pot_abs_res = (ref_potential - potential_model) * potential_mask
                vmax = np.max(np.abs(ref_potential)) / 2.
                im = ax.imshow(pot_abs_res, extent=extent,
                               vmin=-vmax, vmax=vmax,
                               cmap=self.cmap_res)
                ax.set_title(r"$\psi_{\rm pix}$ - $\psi_{\rm pix, ref}$", fontsize=self.base_fontsize)
                nice_colorbar_residuals(im, pot_abs_res, position='top', pad=0.4, size=0.2, 
                                        vmin=-vmax, vmax=vmax,
                                        colorbar_kwargs={'orientation': 'horizontal'})
                im.set_rasterized(True)
                ax.imshow(likelihood_mask_nans, extent=extent, cmap='gray_r', vmin=0, vmax=1)
            else:
                ax.axis('off')
            i_row += 1

        if show_lens_others:

            ##### DEFLECTION ANGLES, SURFACE MASS DENSITY AND MAGNIFICATION #####
            ax = axes[i_row, 0]
            im = ax.imshow(alpha_x * potential_mask, cmap=self.cmap_deriv1, alpha=1, extent=extent)
            im.set_rasterized(True)
            title = r"$\alpha_{x,\rm pix}$" if only_pixelated_potential else r"deflection field ($x$)"
            ax.set_title(title, fontsize=self.base_fontsize)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})
            ax = axes[i_row, 1]
            im = ax.imshow(kappa * potential_mask, cmap=self.cmap_deriv2, norm=LogNorm(),
                           alpha=1, extent=extent)
            im.set_rasterized(True)
            title = r"$\kappa_{\rm pix}$" if only_pixelated_potential else "convergence"
            ax.set_title(title, fontsize=self.base_fontsize)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})
            ax = axes[i_row, 2]
            im = ax.imshow(magnification * potential_mask, cmap=self.cmap_default, norm=Normalize(-10, 10),
                           alpha=1, extent=extent)
            im.set_rasterized(True)
            title = r"$\mu_{\rm pix}$" if only_pixelated_potential else "magnification"
            ax.set_title(title, fontsize=self.base_fontsize)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})
            i_row += 1

        if show_plot:
            plt.show()
        return fig
    

    def model_summary_multiplane(
            self,
            lens_image,
            kwargs_result,
            kwargs_noise=None,
            likelihood_mask=None,
            masked_residuals=True,
            show_shear_field=False,
            norm_kappa=LogNorm(1e-1, 1e1),
            extent_zoom=[-0.5, 0.5, -0.5, 0.5],
            kwargs_grid_source=None,
            linestyles_planes=[':', '-', '--', '-.', ':', '-'],
            colors_planes=['tab:purple', 'tab:cyan', 'tab:orange', 'tab:pink', 'tab:blue', 'tab:red'],
            supersampling_lens_lines=3,
        ):
        """
        Simple function with limited user-control to plot the details
        of a multi-plane lens model based on a MPLensImage instance.

        NOTE: This function will likely improve / be heavily revamped in the future.

        Here `kwargs_grid_source` can also be a list for each source plane
        """
        if not isinstance(lens_image, MPLensImage):
            raise ValueError("This function is only for multi-plane lens models.")
        if kwargs_noise is None:
            kwargs_noise = {}
        
        # get the total number of planes as we will iterate over it
        num_planes = lens_image.MPLightModel.number_light_planes

        # optional grid parameters of sources for each plane
        if not isinstance(kwargs_grid_source, (list, tuple)):
            kwargs_grid_source = [copy.deepcopy(kwargs_grid_source) for _ in range(num_planes - 1)]

        # create a figure with multiple subplots:
        # - the first row contains the data, the model and the normalized residuals, and the model with critical lines overlayed
        # - the second row contains the conjugate points, the convergence map, the (total) shear map as a vector field, and the magnification map
        # - the following rows are for each individual plane, each one showing, in order:
        #   - the data with the light in that plane subtracted
        #   - that light model in the image plane
        #   - that light model in its own plane
        #   - that light model in its own plane with caustics overlayed

        # Create the figure and axes following the same dimensions as the model_summary function
        n_cols = 4
        n_rows = 2 + num_planes
        figsize = (15, n_rows*5)
        fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)

        # Get the data and model images
        if hasattr(self, '_data'):
            data = self._data
        else:
            data = np.zeros(lens_image.Grid.number_pixel_axes)

        # Get the model image
        model = lens_image.model(
            **kwargs_result, 
        )

        # Get the residuals
        model_residuals, residuals = lens_image.normalized_residuals(
            data, model, kwargs_noise=kwargs_noise, mask=likelihood_mask,
        )
        if masked_residuals is True:
            residuals_plot = model_residuals
        else:
            residuals_plot = residuals
        red_chi2 = lens_image.reduced_chi2(
            data, model, kwargs_noise=kwargs_noise, mask=likelihood_mask,
        )
        # Get the extent of the image
        extent = lens_image.Grid.plt_extent
        # Get the critical lines and caustics
        clines_per_plane = []
        caustics_per_plane = []
        centers_per_plane = []
        for idx_plane in range(1, num_planes)[::-1]:
            clines, caustics, centers = model_util.critical_lines_caustics(
                lens_image, 
                kwargs_result['kwargs_mass'], 
                eta_flat=kwargs_result['eta_flat'], 
                return_lens_centers=True,
                k_plane=idx_plane,
                supersampling=supersampling_lens_lines,
            )
            clines_per_plane.append(clines)
            caustics_per_plane.append(caustics)
            centers_per_plane.append(centers)
        # Get the convergence map
        # x_grid, y_grid = lens_image.Grid.pixel_coordinates
        x_grid_highres, y_grid_highres = lens_image.Grid.create_model_grid(
            pixel_scale_factor=0.2
        ).pixel_coordinates
        kappa = lens_image.MPMassModel.kappa(
            x_grid_highres, y_grid_highres, 
            kwargs_result['eta_flat'],
            kwargs_result['kwargs_mass'], 
        )
        kappa = kappa[-1]
        # Get the magnification map
        magnification = lens_image.MPMassModel.magnification(
            x_grid_highres, y_grid_highres, 
            kwargs_result['eta_flat'],
            kwargs_result['kwargs_mass'], 
        )
        magnification = magnification[-1]
        # Get the shear field
        if show_shear_field:
            shear_field = model_util.total_shear_deflection_field(
                lens_image, 
                kwargs_result['kwargs_mass'], 
                eta_flat=kwargs_result['eta_flat'],
                num_pixels=20,
                k_plane=-1,
            )
        else:
            shear_field = None
        # Get the conjugate points
        conj_points_per_plane = lens_image.conjugate_points
        no_conj_points = all([cp is None for cp in conj_points_per_plane])
        traced_conj_points_per_plane = []
        for idx_plane in range(1, num_planes):
            traced_conj_points = lens_image.trace_conjugate_points(
                kwargs_result['eta_flat'],
                kwargs_result['kwargs_mass'], 
                N=idx_plane, k_mass=None,
            )
            traced_conj_points_per_plane.append(traced_conj_points)
        # print("conjugate points:", conj_points_per_plane)
        # print("traced conjugate points:", traced_conj_points_per_plane)

        # We also get the convergence and lens light in plane 0 to compare their isocontours
        # Get the lens light model
        if not lens_image.MPLightModel.light_models[0].has_pixels:
            main_lens_light = lens_image.MPLightModel.light_models[0].surface_brightness( 
                x_grid_highres, y_grid_highres, 
                kwargs_result['kwargs_light'][0], 
                pixels_x_coord=None, pixels_y_coord=None,  # NOTE: this means it would not work with a pixelated ,
            )
        else:
            # TODO: this has not been tested
            main_lens_light = lens_image.model(
                k_light=[[0]],
                **kwargs_result,
            )
            main_lens_light = main_lens_light[0]
        main_lens_mass = lens_image.MPMassModel.mass_models[0].kappa(
            x_grid_highres, y_grid_highres, 
            kwargs_result['kwargs_mass'][0], 
        )
        
        # Populate the figure
        # Data
        ax = axes[0, 0]
        im = ax.imshow(data, extent=extent, cmap=self.cmap_flux, norm=self.norm_flux)
        im.set_rasterized(True)
        ax.set_title("Data", fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                      colorbar_kwargs={'orientation': 'horizontal'})
        
        # Model
        ax = axes[0, 1]
        im = ax.imshow(model, extent=extent, cmap=self.cmap_flux, norm=self.norm_flux)
        im.set_rasterized(True)
        self._plot_masks_contours(
            ax, lens_image, likelihood_mask=likelihood_mask, extent=extent, color='white', alpha=0.4,
        )
        ax.set_title("Model", fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                      colorbar_kwargs={'orientation': 'horizontal'})
        
        # Residuals
        ax = axes[0, 2]
        im = ax.imshow(residuals_plot, cmap=self.cmap_res, extent=extent, norm=self.norm_res)
        im.set_rasterized(True)
        self._plot_masks_contours(
            ax, lens_image, likelihood_mask=likelihood_mask, extent=extent, color='black', alpha=0.6,
        )
        if likelihood_mask is not None and masked_residuals is False:
            ax.contour(likelihood_mask, extent=extent, levels=[0], 
                       colors='black', alpha=0.5, linewidths=0.5)
            
        ax.set_title(r"(f${}_{\rm data}$ - f${}_{\rm model})/\sigma$", fontsize=self.base_fontsize)
        nice_colorbar_residuals(im, residuals_plot, position='top', pad=0.4, size=0.2, 
                                vmin=self.norm_res.vmin, vmax=self.norm_res.vmax,
                                colorbar_kwargs={'orientation': 'horizontal'})
        text = r"$\chi^2_\nu={:.2f}$".format(red_chi2)
        ax.text(0.05, 0.05, text, color='black', fontsize=self.base_fontsize-4, 
                horizontalalignment='left', verticalalignment='bottom',
                transform=ax.transAxes, bbox={'color': 'white', 'alpha': 0.8})
        
        # Model with critical lines overlayed
        ax = axes[0, 3]
        im = ax.imshow(data, extent=extent, cmap=self.cmap_bw, norm=self.norm_flux)
        im.set_rasterized(True)
        ax.set_title("Data + critical lines", fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                      colorbar_kwargs={'orientation': 'horizontal'})
        if likelihood_mask is not None:
            ax.contour(likelihood_mask, extent=extent, levels=[0], 
                       colors='black', alpha=0.5, linewidths=0.5)
        for i, clines in enumerate(clines_per_plane[::-1]):
            src_plane_idx = i + 1
            for j, curve in enumerate(clines):
                ax.plot(curve[0], curve[1], linewidth=2, color=colors_planes[src_plane_idx], linestyle=linestyles_planes[src_plane_idx], 
                        label=f"Source in plane {src_plane_idx}" if j == 0 else None)
        ax.scatter(*centers, s=20, c='gray', marker='+', linewidths=0.5)
        ax.legend()
        # tighten the layout and show the figure
        fig.tight_layout()

        # Conjugate points
        ax = axes[1, 0]
        if not no_conj_points:
            im = ax.imshow(data, extent=extent, cmap=self.cmap_bw, norm=self.norm_flux, 
                           alpha=1)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'},
                          invisible=True)
            self._plot_conjugate_points_in_ax(
                ax, num_planes, 
                conj_points_per_plane, 
                traced_conj_points_per_plane,
                extent, extent_zoom=None,
                colors_planes=colors_planes[1:],
            )
            ax.legend(loc='best', fontsize=8)
            ax.set_title("Data + conjugate points", fontsize=self.base_fontsize)
        else:
            ax.axis('off')
            # ax.set_title("No conjugate points", fontsize=self.base_fontsize)
        
        # Zoom-in on the traced conjuate points, or the shear field
        ax = axes[1, 1]
        if show_shear_field:
            x_field, y_field, gx_field, gy_field = shear_field
            ax.quiver(
                x_field, y_field,
                gx_field, gy_field, 
                scale=3, 
                width=0.1,
                scale_units='xy', 
                units='xy',
                pivot='middle',
                headaxislength=0, headlength=0,
                color='black', alpha=0.3,
            )
            ax.set_aspect('equal')
            ax.set_xlim(extent[0], extent[1])
            ax.set_ylim(extent[2], extent[3])
            ax.set_title(f"Shear field (w.r.t. plane {num_planes-1})", fontsize=self.base_fontsize)
        else:
            im = ax.imshow(data, extent=extent, cmap=self.cmap_bw, norm=self.norm_flux, alpha=0)  # zero opacity
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'},
                          invisible=True)
            self._plot_conjugate_points_in_ax(
                ax, num_planes, 
                conj_points_per_plane, traced_conj_points_per_plane,
                extent, extent_zoom=extent_zoom,
                colors_planes=colors_planes[1:],
            )
            ax.legend(loc='best', fontsize=8)

        # Convergence map
        ax = axes[1, 2]
        im = ax.imshow(kappa, extent=extent, cmap=self.cmap_default, norm=norm_kappa)
        im.set_rasterized(True)
        ax.set_title(r"Convergence $\kappa$ (image plane)", fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                      colorbar_kwargs={'orientation': 'horizontal'})
        # overlay the contours of the lens light and mass models
        contour_with_legend(
            ax,
            [
                main_lens_mass, 
                main_lens_light,
            ], 
            [
                dict(extent=extent, levels=10,
                     colors='black', alpha=0.5,
                     linewidths=0.5, linestyles='-',
                     label="Main lens mass"),
                dict(extent=extent, levels=5,
                     colors='black', alpha=0.5,
                     linewidths=1, linestyles='--',
                     label="Main lens light"),
            ], 
            logscale=True,
        )

        # Magnification map
        ax = axes[1, 3]
        im = ax.imshow(magnification, extent=extent, cmap=self.cmap_default, norm=Normalize(-20, 20), alpha=0.7)
        im.set_rasterized(True)
        ax.set_title(r"Magnification $\mu$ " + f"(w.r.t. plane {num_planes-1})", fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                      colorbar_kwargs={'orientation': 'horizontal'})
        
        # The remaining axes are the same for each single plane that contains light (lens or source) models
        for i, idx_plane in enumerate(range(num_planes)):
            i_row = i + 2
            axes_row = axes[i_row, :]
            kwargs_grid_source_i = kwargs_grid_source[i-1] if i > 1 else None  # skip lens light plane
            self._plot_data_with_single_plane_model(
                axes_row,
                idx_plane,
                data,
                lens_image,
                kwargs_result,
                extent,
                kwargs_grid_source=kwargs_grid_source_i,
                linestyle=linestyles_planes[i],
                color=colors_planes[i],
            )

        fig.tight_layout()
        return fig

    def plot_conjugate_points(
            self,
            mp_lens_image, 
            kwargs_result, 
            num_zooms=1, 
            extent=None,
            extent_zoom=None,
            redshifts_conjugate_points=None,
            ref_source_plane_positions=None,
            index_source_plane_lines=1,
            supersampling_critical_lines=2,
            show_plot=True, 
            fig_width=15,
        ):
        if num_zooms > 1:
            assert len(extent_zoom) == num_zooms
            extent_zoom_list = extent_zoom
        else:
            if len(extent_zoom) == 4:
                extent_zoom_list = [extent_zoom]
            else:
                extent_zoom_list = extent_zoom
        if not isinstance(index_source_plane_lines, (list, tuple)):
            index_source_plane_lines = [index_source_plane_lines]
        # Create the figure and axes following the same dimensions as the model_summary function
        n_cols = 1
        n_rows = 1 + num_zooms
        figsize = (fig_width, fig_width/3.*n_rows)
        fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)
        if n_rows == 1:
            axes = [axes]
        # get the total number of planes as we will iterate over it
        num_planes = mp_lens_image.MPLightModel.number_light_planes
        # get the model extent
        if extent is None:
            extent = mp_lens_image.Grid.plt_extent
        # Get the conjugate points
        conj_points_per_plane = mp_lens_image.conjugate_points
        #no_conj_points = all([cp is None for cp in conj_points_per_plane])
        traced_conj_points_per_plane = []
        for idx_plane in range(1, num_planes):
            traced_conj_points = mp_lens_image.trace_conjugate_points(
                kwargs_result['eta_flat'],
                kwargs_result['kwargs_mass'], 
                N=idx_plane, 
                k_mass=None,
            )
            traced_conj_points_per_plane.append(traced_conj_points)
        clines_per_plane = []
        for idx_plane in index_source_plane_lines:
            clines, _ = model_util.critical_lines_caustics(
                mp_lens_image, 
                kwargs_result['kwargs_mass'], 
                eta_flat=kwargs_result['eta_flat'], 
                return_lens_centers=False,
                k_plane=idx_plane+1,
                supersampling=supersampling_critical_lines,
            )
            clines_per_plane.append((idx_plane, clines))
        # create a large color palette with a variety of colors
        colors_per_planes = mpl.color_sequences['tab20'] + mpl.color_sequences['Pastel1'] + mpl.color_sequences['Pastel2']
        # if the data has been set, also show the data
        data = self._data if hasattr(self, '_data') else None
        for i, ax in enumerate(axes):
            if data is not None:
                im = axes[0].imshow(data, extent=mp_lens_image.Grid.plt_extent, 
                                    cmap=self.cmap_bw, norm=self.norm_flux, alpha=0.8)
                im.set_rasterized(True)
            self._plot_conjugate_points_in_ax(
                ax, num_planes, 
                conj_points_per_plane, 
                traced_conj_points_per_plane,
                extent, 
                extent_zoom=extent_zoom_list[i-1] if i > 0 else None,
                colors_planes=colors_per_planes,
                redshift_per_plane=redshifts_conjugate_points,
                ref_traced_conj_points_per_plane=ref_source_plane_positions,
            )
            for idx_plane, clines in clines_per_plane:
                for j, curve in enumerate(clines):
                    ax.plot(curve[0], curve[1], 
                            color=colors_per_planes[idx_plane],   # -1 because colors_per_planes starts at 0
                            linestyle='--', linewidth=1.5, alpha=1, 
                            label=f"Critical lines for z={redshifts_conjugate_points[idx_plane] if redshifts_conjugate_points is not None else -1.:.3f}" if j == 0 else None)
            ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), fontsize=8)
            if i == 0:
                ax.set_title("Base view")
            else:
                ax.set_title(f"Zoom-in view {i}")
        fig.tight_layout()
        if show_plot:
            plt.show()
        return fig

    def imshow_flux(self, ax, image, colorbar=True):
        im = ax.imshow(image, cmap=self.cmap_flux, norm=self.norm_flux)
        if colorbar is True:
            nice_colorbar(im)


    def _get_norm_for_model(self, model, lock_colorbars):
        if lock_colorbars is True and self.norm_flux is None:
            if self.flux_log_scale is True:
                norm_flux = LogNorm(model.min(), model.max())
            else:
                norm_flux = Normalize(model.min(), model.max())
        else:
            norm_flux = self.norm_flux
        return norm_flux
    
    @staticmethod
    def _plot_masks_contours(ax, lens_image, likelihood_mask=None, extent=None, alpha=0.5, color='red'):
        if likelihood_mask is not None:
            ax.contour(likelihood_mask, extent=extent, levels=[0], 
                    colors=color, alpha=alpha, linewidths=1.5, linestyles='-',
                    label="Likelhood mask")
        if not all([m is None for m in lens_image.source_arc_masks]):
            for j, mask in enumerate(lens_image.source_arc_masks):
                ax.contour(mask, extent=extent, levels=[0], 
                        colors=color, alpha=alpha, linewidths=1.5, linestyles=':',
                        label="Source arc mask" if j == 0 else None)

    @staticmethod
    def _plot_conjugate_points_in_ax(
            ax, num_planes, 
            conj_points_per_plane,
            traced_conj_points_per_plane, 
            extent, 
            extent_zoom=None,
            colors_planes=['tab:orange', 'tab:purple', 'tab:cyan', 'tab:pink', 'tab:blue', 'tab:red'],
            redshift_per_plane=None,
            ref_traced_conj_points_per_plane=None,
        ):
        for idx_plane in range(num_planes-1):
            if conj_points_per_plane[idx_plane] is None:
                continue
            color = colors_planes[idx_plane % len(colors_planes)]
            cps = conj_points_per_plane[idx_plane].T
            tcps = traced_conj_points_per_plane[idx_plane].T
            # check if the conjugate points are in the extent
            extent_check = extent_zoom if extent_zoom is not None else extent
            if not (np.any(
                    ((extent_check[0] < cps[0]) & (cps[0] < extent_check[1])) &
                    ((extent_check[2] < cps[1])) & (cps[1] < extent_check[3])
                ) or np.any(
                    ((extent_check[0] < tcps[0]) & (tcps[0] < extent_check[1])) &
                    ((extent_check[2] < tcps[1])) & (tcps[1] < extent_check[3])
                )):
                continue
            # plot the conjugate points and traced conjugate points
            label = f"Plane {idx_plane+1}"
            if redshift_per_plane is not None:
                label += f" (z={redshift_per_plane[idx_plane]:.3f})"
            ax.scatter(*cps, s=140, edgecolors=color, marker='o', linewidths=2, facecolors='none',
                       label=label)
            ax.scatter(*tcps, s=400 if extent_zoom else 140, facecolors='none', edgecolors=color, marker='*', linewidths=2)
            if (ref_traced_conj_points_per_plane is not None
                and ref_traced_conj_points_per_plane[idx_plane] is not None):
                ref_tcps = ref_traced_conj_points_per_plane[idx_plane].T
                ax.scatter(*ref_tcps, s=400 if extent_zoom else 140, facecolors=color, edgecolors='none', marker='*', linewidths=0)
            for conj_point, traced_conj_point in zip(conj_points_per_plane[idx_plane], traced_conj_points_per_plane[idx_plane]):
                conj_point_arrow = FancyArrowPatch(
                    (traced_conj_point[0], traced_conj_point[1]), 
                    (conj_point[0], conj_point[1]), 
                    arrowstyle=ArrowStyle("Fancy", head_length=1, head_width=1, tail_width=1), 
                    color=color, lw=2, alpha=0.2, clip_on=True,
                )
                ax.add_patch(conj_point_arrow)
        ax.set_aspect('equal')
        lim_extent = extent_zoom if extent_zoom is not None else extent
        ax.set_xlim(lim_extent[0], lim_extent[1])
        ax.set_ylim(lim_extent[2], lim_extent[3])
        
    def _plot_data_with_single_plane_model(
        self,
        axes_row,
        k_plane,
        data,
        lens_image,
        kwargs_result,
        extent,
        kwargs_grid_source=None,
        linestyle='-',
        color='tab:blue',
        ):
        """Populate axes (single row) with, in order:
           - the data with the light in that plane subtracted
           - that light model in the image plane
           - that light model in its own plane
           - that light model in its own plane with caustics overlayed
        """
        # Get the lensed source model of the required plane
        observed_light_model = lens_image.model(
            k_planes=k_plane,
            **kwargs_result,
        )
        # Subtract the lensed source model from the data
        data_subtracted = data - observed_light_model
        # Get the unlensed source model of the required plane
        if k_plane > 0:
            plane_has_pixels = lens_image.MPLightModel.light_models[k_plane].has_pixels
            if plane_has_pixels:
                # TODO: the following complicated code should be implemented wthin MPLensImage (similar to LensImage)
                # get the adapted coordinates axes
                x_coord, y_coord, orig_extent_tmp = lens_image.get_source_coordinates(
                    kwargs_result['eta_flat'],
                    kwargs_result['kwargs_mass'],
                )
                # select the right plane
                x_coord, y_coord = x_coord[k_plane], y_coord[k_plane]
                orig_extent_tmp = orig_extent_tmp[k_plane]  # NOTE: this is not a "plot" extent (misses half pixels at each end)
            else:
                x_coord, y_coord = None, None
            if kwargs_grid_source is not None:
                grid_src = lens_image.Grid.create_model_grid(**kwargs_grid_source)
                x_grid_src, y_grid_src = grid_src.pixel_coordinates
                orig_extent = grid_src.plt_extent
            elif plane_has_pixels:
                # create a 2d grid out of the axes
                x_grid_src, y_grid_src = np.meshgrid(x_coord, y_coord)
                # transform the proper extent to the one usable with imshow 
                pix_scl_x = jnp.abs(orig_extent_tmp[0]-orig_extent_tmp[1])
                pix_scl_y = jnp.abs(orig_extent_tmp[2]-orig_extent_tmp[3])
                half_pix_scl = jnp.sqrt(pix_scl_x*pix_scl_y) / 2.
                orig_extent = [
                    orig_extent_tmp[0]-half_pix_scl, orig_extent_tmp[1]+half_pix_scl, 
                    orig_extent_tmp[2]-half_pix_scl, orig_extent_tmp[3]+half_pix_scl
                ]
            else:
                # fall-back case when no grid is provided nor found
                x_grid_src, y_grid_src = lens_image.ImageNumerics.coordinates_evaluate  # NOTE: coordinates are 1d here
                orig_extent = extent
            original_light_model = lens_image.MPLightModel.light_models[k_plane].surface_brightness(
                x_grid_src, y_grid_src, 
                kwargs_result['kwargs_light'][k_plane],
                pixels_x_coord=x_coord,
                pixels_y_coord=y_coord,
            ) * lens_image.Grid.pixel_area
            # if original_light_model.ndim == 1:
            #     original_light_model = lens_image.ImageNumerics.re_size_convolve(
            #         original_light_model, unconvolved=True, input_as_list=False,
            #     )
        else:
            # for the lens plane, we do not have a source model
            # original_light_model = lens_image.model(
            #     k_planes=k_plane,
            #     unconvolved=True,
            #     **kwargs_result,
            # )
            # orig_extent = extent
            x_grid_model, y_grid_model = lens_image.ImageNumerics.coordinates_evaluate  # NOTE: coordinates are 1d here
            original_light_model = lens_image.MPLightModel.light_models[0].surface_brightness(
                x_grid_model, y_grid_model, 
                kwargs_result['kwargs_light'][0],
                pixels_x_coord=None,
                pixels_y_coord=None,
            ) * lens_image.Grid.pixel_area
            orig_extent = extent
        
        # typically when we used lens_image.ImageNumerics.coordinates_evaluate
        if original_light_model.ndim == 1:
            n_side = int(np.sqrt(original_light_model.size))
            original_light_model = original_light_model.reshape((n_side, n_side))

        # Get the caustics
        _, caustics = model_util.critical_lines_caustics(
            lens_image, 
            kwargs_result['kwargs_mass'], 
            eta_flat=kwargs_result['eta_flat'], 
            return_lens_centers=False,
            k_plane=k_plane,
            supersampling=6,
        )

        # Get the traced conjugate points
        traced_conj_points = lens_image.trace_conjugate_points(
            kwargs_result['eta_flat'],
            kwargs_result['kwargs_mass'], 
            N=k_plane,
        )

        # Source-subtracted data
        ax = axes_row[0]
        im = ax.imshow(data_subtracted, extent=extent, cmap=self.cmap_flux, norm=self.norm_flux)
        im.set_rasterized(True)
        ax.set_title(f"Subtracted model in plane {k_plane}", fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                      colorbar_kwargs={'orientation': 'horizontal'})
        
        # Light model in image plane
        ax = axes_row[1]
        im = ax.imshow(observed_light_model, extent=extent, cmap=self.cmap_flux, norm=self.norm_flux)
        im.set_rasterized(True)
        ax.set_title(f"Obs. light model in plane {k_plane}", fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                      colorbar_kwargs={'orientation': 'horizontal'})
        
        # Light model in "original form"
        ax = axes_row[2]
        im = ax.imshow(original_light_model, extent=orig_extent, cmap=self.cmap_flux_alt, norm=self.norm_flux)
        im.set_rasterized(True)
        if k_plane > 0:
            ax.set_title(f"Unlensed model in plane {k_plane}", fontsize=self.base_fontsize)
        else:
            ax.set_title(f"Unconvolved model in plane {k_plane}", fontsize=self.base_fontsize)
        nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                        colorbar_kwargs={'orientation': 'horizontal'})
            
        # Source model with caustics overlayed )
        ax = axes_row[3]
        if k_plane > 0:
            im = ax.imshow(original_light_model, extent=orig_extent, cmap=self.cmap_bw, norm=self.norm_flux)
            im.set_rasterized(True)
            nice_colorbar(im, position='top', pad=0.4, size=0.2, 
                          colorbar_kwargs={'orientation': 'horizontal'})
            ax.set_title(f"With caustics + conj. points", fontsize=self.base_fontsize)
            # FIXME: caustics are not plotted correctly
            for j, curve in enumerate(caustics):
                # print(j, curve)
                ax.plot(curve[0], curve[1], linewidth=2, color=color, linestyle=linestyle)
            ax.scatter(*traced_conj_points.T, c=color, s=300, marker='*', linewidths=1, edgecolors='white', facecolors=color)
            ax.set_xlim(orig_extent[0], orig_extent[1])
            ax.set_ylim(orig_extent[2], orig_extent[3])
            ax.set_aspect('equal')
        else:
            ax.axis('off')



# Defines a singular isothermal sphere
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LensModel.Profiles module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
from herculens.MassModel.Profiles.sie import SIE


__all__ = ['SIS']


class SIS(object):
    """Singular isothermal sphere mass profile."""
    param_names = ['theta_E', 'center_x', 'center_y']
    lower_limit_default = {'theta_E': 0, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'theta_E': 100, 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}
    
    def __init__(self):
        self.profile = SIE()
        self._e = 1e-4
        super(SIS, self).__init__()

    def function(self, x, y, theta_E, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param theta_E:
        :param center_x:
        :param center_y:
        :return:
        """
        return self.profile.function(x, y, theta_E, self._e, self._e, center_x, center_y)

    def derivatives(self, x, y, theta_E, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param theta_E:
        :param center_x:
        :param center_y:
        :return:
        """
        return self.profile.derivatives(x, y, theta_E, self._e, self._e, center_x, center_y)

    def hessian(self, x, y, theta_E, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param theta_E:
        :param center_x:
        :param center_y:
        :return:
        """
        return self.profile.hessian(x, y, theta_E, self._e, self._e, center_x, center_y)

    @staticmethod
    def theta2rho(theta_E):
        """
        converts projected density parameter (in units of deflection) into 3d density parameter
        :param theta_E:
        :return:
        """
        self.profile.theta2rho(theta_E)

    @staticmethod
    def mass_3d(r, rho0):
        """
        mass enclosed a 3d sphere or radius r
        :param r: radius in angular units
        :param rho0: density at angle=1
        :return: mass in angular units
        """
        self.profile.mass_3d(r, rho0)

    def mass_3d_lens(self, r, theta_E):
        """
        mass enclosed a 3d sphere or radius r given a lens parameterization with angular units

        :param r: radius in angular units
        :param theta_E: Einstein radius
        :return: mass in angular units
        """
        return self.profile.mass_3d_lens(r, theta_E)

    def mass_2d(self, r, rho0):
        """
        mass enclosed projected 2d sphere of radius r
        :param r:
        :param rho0:
        :param a:
        :param s:
        :return:
        """
        return self.profile.mass_2d(r, rho0)

    def mass_2d_lens(self, r, theta_E):
        """

        :param r:
        :param theta_E:
        :return:
        """
        return self.profile.mass_2d_lens(r, theta_E)

    def grav_pot(self, x, y, rho0, center_x=0, center_y=0):
        """
        gravitational potential (modulo 4 pi G and rho0 in appropriate units)
        :param x:
        :param y:
        :param rho0:
        :param a:
        :param s:
        :param center_x:
        :param center_y:
        :return:
        """
        return self.profile.grav_pot(x, y, rho0, center_x=center_x, center_y=center_y)

    def density_lens(self, r, theta_E):
        """
        computes the density at 3d radius r given lens model parameterization.
        The integral in the LOS projection of this quantity results in the convergence quantity.

        :param r: radius in angles
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :return: density
        """
        return self.profile.density_lens(r, theta_E)

    def density(self, r, rho0):
        """
        computes the density
        :param r: radius in angles
        :param rho0: density at angle=1
        :return: density at r
        """
        return self.profile.density(rho0)

    def density_2d(self, x, y, rho0, center_x=0, center_y=0):
        """
        projected density
        :param x:
        :param y:
        :param rho0:
        :param center_x:
        :param center_y:
        :return:
        """
        return self.profile.density_2d(x, y, rho0, center_x=center_x, center_y=center_y)



# Handles coordinate systems
# 
# Copyright (c) 2022, herculens developers and contributors


__author__ =  'aymgal'


import os
from copy import deepcopy

import herculens
from herculens.Standard import coolest_util as util

from coolest.template.json import JSONSerializer
from coolest.api.util import get_coolest_object


class COOLESTexporter(object):
    """Class that handles conversion from a Herculens model to the COOLEST file system"""

    def __init__(self, output_basename, output_directory, input_coolest_file=None, 
                 empty_output_directory=False, **kwargs_serializer):
        if input_coolest_file is None:
            raise NotImplementedError("You must provide an input coolest file (for now)")
        if not os.path.isabs(input_coolest_file):
            self._input_coolest_file = os.path.abspath(input_coolest_file)
        else:
            self._input_coolest_file = input_coolest_file
        output_coolest_file = os.path.join(output_directory, output_basename)
        if not os.path.isabs(output_coolest_file):
            self._output_coolest_file = os.path.abspath(output_coolest_file)
        else:
            self._output_coolest_file = output_coolest_file
        self._output_dir = os.path.dirname(output_coolest_file)
        self._basename = output_basename
        self._kwargs_serializer = kwargs_serializer
        self._coolest = self._load_coolest_object()
        dir_bool = util.create_directory(self._output_dir, empty_output_directory)
        if dir_bool is False:
            print("COOLEST-warning: Output directory already exists and has not been emptied. "
                  "The template file might be directly updated.")

    @property
    def coolest_object(self):
        return self._coolest

    def save_on_disk(self, suffix=None, with_jsonpickle=False):
        # TODO: save a directory instead of just a json file
        output_file = self._output_coolest_file
        if suffix is not None:
            output_file += '-' + suffix
        serializer = JSONSerializer(output_file, obj=self._coolest, 
                                   **self._kwargs_serializer)
        if with_jsonpickle:
            serializer.dump_jsonpickle()
        else:
            serializer.dump_simple()
        print("COOLEST-info: successfully saved the updated template file.")

    def update_from_data(self, data, lens_image,
                         noise_type='NoiseMap', noise_map=None,
                         psf_type='PixelatedPSF', psf_description=None,
                         kwargs_obs=None, kwargs_noise=None, kwargs_instrument=None):
        if 'mag_zero_point' not in kwargs_obs:
            kwargs_obs['mag_zero_point'] = self._coolest.observation.mag_zero_point
        if 'mag_sky_brightness' not in kwargs_obs:
            kwargs_obs['mag_sky_brightness'] = self._coolest.observation.mag_sky_brightness
        observation = util.create_observation(data, lens_image, 
                                              noise_type=noise_type,
                                              model_noise_map=noise_map, 
                                              json_dir=self._output_dir,
                                              kwargs_obs=kwargs_obs,
                                              kwargs_noise=kwargs_noise)
        instrument = util.create_instrument(lens_image, observation,
                                            json_dir=self._output_dir,
                                            psf_type=psf_type,
                                            psf_description=psf_description,
                                            kwargs=kwargs_instrument)
        # overwites the attributes
        self._coolest.observation = observation
        self._coolest.instrument = instrument

    def update_from_model(self, lens_image, lensing_entity_mapping, 
                          parameters=None, samples=None, 
                          re_create_entities=False):
        lensing_entities = util.update_lensing_entities(lens_image, 
                                                        deepcopy(lensing_entity_mapping),
                                                        parameters=parameters, samples=samples,
                                                        re_create_entities=re_create_entities,
                                                        current_entities=self._coolest.lensing_entities,
                                                        json_dir=self._output_dir,
                                                        fits_file_suffix="bestfit")
        if re_create_entities is True:
            # overwites the lensing entities of the COOLEST object
            self._coolest.lensing_entities = lensing_entities
        if parameters is not None or samples is not None:
            # set the COOLEST mode to MAP (i.e., maximum a-posteriori estimate)
            self._coolest.mode = "MAP"

    def update_from_likelihood(self, lens_image,
                               likelihood_type="ImagingDataLikelihood",
                               likelihood_mask=None):
        likelihoods = util.create_likelihoods(
            lens_image, likelihood_type=likelihood_type,
            likelihood_mask=likelihood_mask,
            json_dir=self._output_dir,
        )
        self._coolest.likelihoods = likelihoods

    def update_from_wcs_coordinates(self, skycoord):
        self._coolest.coordinates_origin = util.skycoord_to_coolest(skycoord)

    def update_metadata(self, **meta_kwargs):
        self._coolest.meta['modeling_code'] = f"Herculens (v{herculens.__version__})"
        self._coolest.meta.update(meta_kwargs)

    def delete_metadata(self, meta_key):
        res = self._coolest.meta.pop(meta_key, None)
        if res is None:
            print(f"COOLEST-warning: key '{meta_key}' has not been found in the COOLEST metadata.")
        else:
            print(f"COOLEST-info: successfully removed key '{meta_key}' from COOLEST metadata.")

    def _load_coolest_object(self):
        coolest_obj = get_coolest_object(self._input_coolest_file, **self._kwargs_serializer)
        return coolest_obj



# Defines the point spread function model
#
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the Data module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import warnings
import numpy as np
from herculens.Util import util, kernel_util
from utax.convolution.functions import build_convolution_matrix


__all__ = ['PSF']


class PSF(object):
    """Point Spread Function class.

    This class describes and manages products used to perform the PSF modeling,
    including convolution for extended surface brightness and painting on the
    PSF for point sources.

    """

    def __init__(self, psf_type='NONE', fwhm=None, truncation=5,
                 pixel_size=None, kernel_point_source=None,
                 kernel_supersampling_factor=1,
                 variance_boost_map=None):
        """Create a PSF object.

        Parameters
        ----------
        psf_type : str, one of {'NONE', 'PIXEL', 'GAUSSIAN'}
            Type of PSF model. Default is 'NONE'.
        fwhm : float, optional
            Full width at half maximum, only required for 'GAUSSIAN' type.
            Default is 'NONE'.
        truncation : float, optional
            Truncation length (in units of sigma) for Gaussian model. Only
            required for 'GAUSSIAN' type. Default is None.
        pixel_size : float, optional
            Pixel width (in arcsec). Required for 'GAUSSIAN' type. Default is None.
        kernel_point_source : array_like, optional
            2D array of odd length representing the centered PSF. Required
            for 'PIXEL' type. Default is None.
        kernel_supersampling_factor : int, optional
            For a 'PIXEL' type PSF, this parameter specifies the factor by
            which the provided `kernel_point_source` has been supersampled.
            Default is 1.
        variance_boost_map : array_like or float, optional
            Single floating point value or 2D array of floating point values 
            that multiply the model noise variance associated to a given pixel value.
            For the 'GAUSSIAN' model, only the single value option is currently supported.
        """
        self.psf_type = psf_type
        self._pixel_size = pixel_size
        if variance_boost_map is None:
            variance_boost_map = 1.

        if self.psf_type == 'GAUSSIAN':
            # Validate required inputs
            if fwhm is None:
                raise ValueError("Must set `fwhm` if `psf_type='GAUSSIAN'`")
            if pixel_size is None:
                raise ValueError("Must set `pixel_size` if `psf_type='GAUSSIAN'`")

            self._fwhm = fwhm
            self._sigma_gaussian = util.fwhm2sigma(self._fwhm)
            self._truncation = truncation
            self._kernel_supersampling_factor = 0
            kernel = self.compute_gaussian_kernel(self._pixel_size, self.fwhm,
                                                  self._truncation)
            self._kernel_point_source = kernel
            self._var_boost_map = variance_boost_map

        elif self.psf_type == 'PIXEL':
            # Validate required inputs
            if kernel_point_source is None:
                raise ValueError(
                    'Must set `kernel_point_source` for PIXEL `psf_type`')
            if len(kernel_point_source) % 2 == 0:
                raise ValueError(
                    'kernel needs to have odd axis number, not ', np.shape(kernel_point_source))
            self._kernel_supersampling_factor = kernel_supersampling_factor
            if kernel_supersampling_factor > 1:
                self._kernel_point_source_supersampled = kernel_point_source
                kernel_point_source = kernel_util.degrade_kernel(
                    self._kernel_point_source_supersampled, self._kernel_supersampling_factor)
            self._kernel_point_source = kernel_point_source / \
                np.sum(kernel_point_source)
            if isinstance(variance_boost_map, (float, int)):
                variance_boost_map = np.full_like(
                    self._kernel_point_source.shape,
                    variance_boost_map,
                )
            elif variance_boost_map.shape != self._kernel_point_source.shape:
                raise ValueError("Variance boost map should have the same shape as the PSF kernel.")
            self._var_boost_map = variance_boost_map
            
        elif self.psf_type == 'NONE':
            self._kernel_point_source = np.zeros((3, 3))
            self._kernel_point_source[1, 1] = 1
            self._var_boost_map = variance_boost_map
    
        else:
            raise ValueError("psf_type %s not supported" % self.psf_type)

    @property
    def kernel_point_source(self):
        if not hasattr(self, '_kernel_point_source'):
            return None
        return self._kernel_point_source

    @property
    def kernel_supersampling_factor(self):
        return self._kernel_supersampling_factor

    @property
    def variance_boost_map(self):
        if not hasattr(self, '_var_boost_map'):
            return None
        return self._var_boost_map

    def kernel_point_source_supersampled(self, supersampling_factor,
                                         iterative_supersampling=False,
                                         update_cache=False):
        """Generate a supersampled PSF.

        The resulting 2D array has the supersampled PSF at the center of a grid
        with odd dimension.

        Parameters
        ----------
        supersampling_factor : int
            Supersampling factor relative to the pixel size.
        iterative_supersampling : bool
            If True, use 5 iterations when supersampling.
        update_cache : bool
            If True, update (and overwrite, if present) the cached supersampled PSF.

        """
        if (hasattr(self, '_kernel_point_source_supersampled') and
                self._kernel_supersampling_factor == supersampling_factor and 
                not update_cache):
            return self._kernel_point_source_supersampled

        if self.psf_type == 'GAUSSIAN':
            npix = self._truncation / self._pixel_size * supersampling_factor
            npix = int(round(npix))
            # Ensure an odd number of pixels
            npix += 1 - npix % 2
            pixel_size = self._pixel_size / supersampling_factor
            result = kernel_util.kernel_gaussian(npix, pixel_size, self._fwhm)
        elif self.psf_type == 'PIXEL':
            num_iter = 5 if iterative_supersampling else 0
            kernel = kernel_util.subgrid_kernel(self.kernel_point_source,
                                                supersampling_factor, odd=True, num_iter=num_iter)
            npix = len(self.kernel_point_source) * supersampling_factor
            npix -= (1 - npix % 2)
            if hasattr(self, '_kernel_point_source_supersampled'):
                warnings.warn("Overwriting supersampled point source kernel " +
                              "due to different subsampling size.")
            result = kernel_util.cut_psf(kernel, psf_size=npix)
        elif self.psf_type == 'NONE':
            result = self._kernel_point_source
        else:
            raise ValueError(f'psf_type {self.psf_type} not valid.')

        if update_cache:
            self._kernel_point_source_supersampled = result
            self._kernel_supersampling_factor = supersampling_factor

        return result
    
    def blurring_matrix(self, data_shape):
        num_pixels = data_shape[0] * data_shape[1]
        if not hasattr(self, '_blurring_matrix') or self._blurring_matrix.shape != (num_pixels, num_pixels):
            psf_kernel_2d = np.array(self.kernel_point_source)
            self._blurring_matrix = build_convolution_matrix(
                psf_kernel_2d, data_shape)
        return self._blurring_matrix

    def set_pixel_size(self, pixel_size):
        """Update pixel size.

        Parameters
        ----------
        pixel_size : float
            New pixel size in angular units (arc seconds).

        Notes
        -----
        The `kernel_point_source` attribute is (re)computed according to the
        new pixel size if the PSF type is GAUSSIAN.

        """
        self._pixel_size = pixel_size

        if self.psf_type == 'GAUSSIAN':
            kernel = self.compute_gaussian_kernel(self._pixel_size, self.fwhm,
                                                  self._truncation)
            self._kernel_point_source = kernel

    def compute_gaussian_kernel(self, pixel_size, fwhm, truncation):
        """Compute a Gaussian kernel matrix to serve as PSF.

        Parameters
        ----------
        pixel_size : float
            Pixel size in angular units (arc seconds).
        fwhm : float
            Full width at half maximum of the Gaussian in pixel units.
        truncation : float
            Truncation length (in units of sigma) for the Gaussian.

        Returns
        -------
        out : array
            2D Gaussian kernel matrix.

        """
        # Determine the number of pixels per side
        npix = round(self._truncation * self.fwhm / self._pixel_size)
        # Ensure an odd number
        npix += 1 - npix % 2
        # Evaluate the 2D Gaussian at the pixel positions
        return kernel_util.kernel_gaussian(npix, self._pixel_size, self.fwhm)

    @property
    def fwhm(self):
        """Full width at half maximum of a Gaussian kernel in pixel units."""
        return self._fwhm
